# pylint: disable=unsupported-assignment-operation,unsubscriptable-object
from __future__ import annotations

import os
import random
from collections.abc import Callable
from typing import Any

import albumentations
import numpy as np
import pandas as pd
import torch
from sklearn.model_selection import train_test_split
from skmultilearn.model_selection import iterative_train_test_split
from timm.data.readers.reader_image_folder import find_images_and_targets
from torch.utils.data import DataLoader

from quadra.datamodules.base import BaseDataModule
from quadra.datasets import ImageClassificationListDataset
from quadra.datasets.classification import MultilabelClassificationDataset
from quadra.utils import utils
from quadra.utils.classification import find_test_image, get_split, group_labels, natural_key

log = utils.get_logger(__name__)


class ClassificationDataModule(BaseDataModule):
    """Base class single folder based classification datamodules. If there is no nested folders, use this class.

    Args:
        data_path: Path to the data main folder.
        name: The name for the data module. Defaults to "classification_datamodule".
        num_workers: Number of workers for dataloaders. Defaults to 16.
        batch_size: Batch size. Defaults to 32.
        seed: Random generator seed. Defaults to 42.
        dataset: Dataset class.
        val_size: The validation split. Defaults to 0.2.
        test_size: The test split. Defaults to 0.2.
        exclude_filter: The filter for excluding folders. Defaults to None.
        include_filter: The filter for including folders. Defaults to None.
        label_map: The mapping for labels. Defaults to None.
        num_data_class: The number of samples per class. Defaults to None.
        train_transform: Transformations for train dataset.
            Defaults to None.
        val_transform: Transformations for validation dataset.
            Defaults to None.
        test_transform: Transformations for test dataset.
            Defaults to None.
        train_split_file: The file with train split. Defaults to None.
        val_split_file: The file with validation split. Defaults to None.
        test_split_file: The file with test split. Defaults to None.
        class_to_idx: The mapping from class name to index. Defaults to None.
        **kwargs: Additional arguments for BaseDataModule.
    """

    def __init__(
        self,
        data_path: str,
        dataset: type[ImageClassificationListDataset] = ImageClassificationListDataset,
        name: str = "classification_datamodule",
        num_workers: int = 8,
        batch_size: int = 32,
        seed: int = 42,
        val_size: float | None = 0.2,
        test_size: float = 0.2,
        num_data_class: int | None = None,
        exclude_filter: list[str] | None = None,
        include_filter: list[str] | None = None,
        label_map: dict[str, Any] | None = None,
        load_aug_images: bool = False,
        aug_name: str | None = None,
        n_aug_to_take: int | None = 4,
        replace_str_from: str | None = None,
        replace_str_to: str | None = None,
        train_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        train_split_file: str | None = None,
        test_split_file: str | None = None,
        val_split_file: str | None = None,
        class_to_idx: dict[str, int] | None = None,
        **kwargs: Any,
    ):
        super().__init__(
            data_path=data_path,
            name=name,
            seed=seed,
            batch_size=batch_size,
            num_workers=num_workers,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            load_aug_images=load_aug_images,
            aug_name=aug_name,
            n_aug_to_take=n_aug_to_take,
            replace_str_from=replace_str_from,
            replace_str_to=replace_str_to,
            **kwargs,
        )
        self.replace_str = None
        self.exclude_filter = exclude_filter
        self.include_filter = include_filter
        self.val_size = val_size
        self.test_size = test_size
        self.label_map = label_map
        self.num_data_class = num_data_class
        self.dataset = dataset
        self.train_split_file = train_split_file
        self.test_split_file = test_split_file
        self.val_split_file = val_split_file
        self.class_to_idx: dict[str, int] | None

        if class_to_idx is not None:
            self.class_to_idx = class_to_idx
            self.num_classes = len(self.class_to_idx)
        else:
            self.class_to_idx = self._find_classes_from_data_path(self.data_path)
            if self.class_to_idx is None:
                log.warning("Could not build a class_to_idx from the data_path subdirectories")
                self.num_classes = 0
            else:
                self.num_classes = len(self.class_to_idx)

    def _read_split(self, split_file: str) -> tuple[list[str], list[str]]:
        """Reads split file.

        Args:
            split_file: Path to the split file.

        Returns:
            List of paths to images.
        """
        samples, targets = [], []
        with open(split_file) as f:
            split = f.readlines()
        for row in split:
            csv_values = row.split(",")
            sample = str(",".join(csv_values[:-1])).strip()
            target = csv_values[-1].strip()
            sample_path = os.path.join(self.data_path, sample)
            if os.path.exists(sample_path):
                samples.append(sample_path)
                targets.append(target)
            else:
                continue
                # log.warning(f"{sample_path} does not exist")
        return samples, targets

    def _find_classes_from_data_path(self, data_path: str) -> dict[str, int] | None:
        """Given a data_path, build a random class_to_idx from the subdirectories.

        Args:
            data_path: Path to the data main folder.

        Returns:
            class_to_idx dictionary.
        """
        subdirectories = []

        # Check if the directory exists
        if os.path.exists(data_path) and os.path.isdir(data_path):
            # Iterate through the items in the directory
            for item in os.listdir(data_path):
                item_path = os.path.join(data_path, item)

                # Check if it's a directory and not starting with "."
                if (
                    os.path.isdir(item_path)
                    and not item.startswith(".")
                    # Check if there's at least one image file in the subdirectory
                    and any(
                        os.path.splitext(file)[1].lower().endswith(tuple(utils.IMAGE_EXTENSIONS))
                        for file in os.listdir(item_path)
                    )
                ):
                    subdirectories.append(item)

            if len(subdirectories) > 0:
                return {cl: idx for idx, cl in enumerate(sorted(subdirectories))}
            return None

        return None

    @staticmethod
    def _find_images_and_targets(
        root_folder: str, class_to_idx: dict[str, int] | None = None
    ) -> tuple[list[tuple[str, int]], dict[str, int]]:
        """Collects the samples from item folders."""
        images_and_targets, class_to_idx = find_images_and_targets(
            folder=root_folder, types=utils.IMAGE_EXTENSIONS, class_to_idx=class_to_idx
        )
        return images_and_targets, class_to_idx

    def _filter_images_and_targets(
        self, images_and_targets: list[tuple[str, int]], class_to_idx: dict[str, int]
    ) -> tuple[list[str], list[str]]:
        """Filters the images and targets."""
        samples: list[str] = []
        targets: list[str] = []
        idx_to_class = {v: k for k, v in class_to_idx.items()}
        images_and_targets = [(str(image_path), target) for image_path, target in images_and_targets]
        for image_path, target in images_and_targets:
            target_class = idx_to_class[target]
            if self.exclude_filter is not None and any(
                exclude_filter in image_path for exclude_filter in self.exclude_filter
            ):
                continue
            if self.include_filter is not None:
                if any(include_filter in image_path for include_filter in self.include_filter):
                    samples.append(str(image_path))
                    targets.append(target_class)
            else:
                samples.append(str(image_path))
                targets.append(target_class)
        return (
            samples,
            targets,
        )

    def _prepare_data(self) -> None:
        """Prepares Classification data for the data module."""
        images_and_targets, class_to_idx = self._find_images_and_targets(self.data_path, self.class_to_idx)
        all_samples, all_targets = self._filter_images_and_targets(images_and_targets, class_to_idx)
        if self.label_map is not None:
            all_targets, _ = group_labels(all_targets, self.label_map)

        samples_train: list[str] = []
        targets_train: list[str] = []
        samples_test: list[str] = []
        targets_test: list[str] = []
        samples_val: list[str] = []
        targets_val: list[str] = []

        if self.test_size < 1.0:
            samples_train, samples_test, targets_train, targets_test = train_test_split(
                all_samples,
                all_targets,
                test_size=self.test_size,
                random_state=self.seed,
                stratify=all_targets,
            )
            if self.test_split_file:
                samples_test, targets_test = self._read_split(self.test_split_file)
                if not self.train_split_file:
                    samples_train, targets_train = [], []
                    for sample, target in zip(all_samples, all_targets, strict=False):
                        if sample not in samples_test:
                            samples_train.append(sample)
                            targets_train.append(target)
            if self.train_split_file:
                samples_train, targets_train = self._read_split(self.train_split_file)
                if not self.test_split_file:
                    samples_test, targets_test = [], []
                    for sample, target in zip(all_samples, all_targets, strict=False):
                        if sample not in samples_train:
                            samples_test.append(sample)
                            targets_test.append(target)
            if self.val_split_file:
                samples_val, targets_val = self._read_split(self.val_split_file)
                if not self.test_split_file or not self.train_split_file:
                    raise ValueError("Validation split file is specified but no train or test split file is specified.")
            else:
                samples_train, samples_val, targets_train, targets_val = train_test_split(
                    samples_train,
                    targets_train,
                    test_size=self.val_size,
                    random_state=self.seed,
                    stratify=targets_train,
                )

            if self.num_data_class is not None:
                samples_train_topick = []
                targets_train_topick = []
                for cl in np.unique(targets_train):
                    idx = np.where(np.array(targets_train) == cl)[0]
                    random.seed(self.seed)
                    random.shuffle(idx)  # type: ignore[arg-type]
                    to_pick = idx[: self.num_data_class]
                    for i in to_pick:
                        samples_train_topick.append(samples_train[i])
                        targets_train_topick.append(cl)

                samples_train = samples_train_topick
                targets_train = targets_train_topick
        else:
            log.info("Test size is set to 1.0: all samples will be put in test-set")
            samples_test = all_samples
            targets_test = all_targets
        train_df = pd.DataFrame({"samples": samples_train, "targets": targets_train})
        train_df["split"] = "train"
        val_df = pd.DataFrame({"samples": samples_val, "targets": targets_val})
        val_df["split"] = "val"
        test_df = pd.DataFrame({"samples": samples_test, "targets": targets_test})
        test_df["split"] = "test"
        self.data = pd.concat([train_df, val_df, test_df], axis=0)

        # if self.load_aug_images:
        #    samples_train, targets_train = self.load_augmented_samples(
        #         samples_train, targets_train, self.replace_str, shuffle=True
        #     )
        #     samples_val, targets_val = self.load_augmented_samples(
        #         samples_val, targets_val , self.replace_str, shuffle=True
        #     )
        unique_targets = [str(t) for t in np.unique(targets_train)]
        if self.class_to_idx is None:
            sorted_targets = sorted(unique_targets, key=natural_key)
            class_to_idx = {c: idx for idx, c in enumerate(sorted_targets)}
            self.class_to_idx = class_to_idx
            log.info("Class_to_idx not provided in config, building it from targets: %s", class_to_idx)

        if len(unique_targets) == 0:
            log.warning("Unique_targets length is 0, training set is empty")
        else:
            if len(self.class_to_idx.keys()) != len(unique_targets):
                raise ValueError(
                    "The number of classes in the class_to_idx dictionary does not match the number of unique targets."
                    f" `class_to_idx`: {self.class_to_idx}, `unique_targets`: {unique_targets}"
                )
            if not all(c in unique_targets for c in self.class_to_idx):
                raise ValueError(
                    "The classes in the class_to_idx dictionary do not match the available unique targets in the"
                    " datasset. `class_to_idx`: {self.class_to_idx}, `unique_targets`: {unique_targets}"
                )

    def setup(self, stage: str | None = None) -> None:
        """Setup data module based on stages of training."""
        if stage in ["train", "fit"]:
            self.train_dataset = self.dataset(
                samples=self.data[self.data["split"] == "train"]["samples"].tolist(),
                targets=self.data[self.data["split"] == "train"]["targets"].tolist(),
                transform=self.train_transform,
                class_to_idx=self.class_to_idx,
            )
            self.val_dataset = self.dataset(
                samples=self.data[self.data["split"] == "val"]["samples"].tolist(),
                targets=self.data[self.data["split"] == "val"]["targets"].tolist(),
                transform=self.val_transform,
                class_to_idx=self.class_to_idx,
            )
        if stage in ["test", "predict"]:
            self.test_dataset = self.dataset(
                samples=self.data[self.data["split"] == "test"]["samples"].tolist(),
                targets=self.data[self.data["split"] == "test"]["targets"].tolist(),
                transform=self.test_transform,
                class_to_idx=self.class_to_idx,
            )

    def train_dataloader(self) -> DataLoader:
        """Returns the train dataloader.

        Raises:
            ValueError: If train dataset is not initialized.

        Returns:
            Train dataloader.
        """
        if not self.train_dataset_available:
            raise ValueError("Train dataset is not initialized")
        if not isinstance(self.train_dataset, torch.utils.data.Dataset):
            raise ValueError("Train dataset has to be single `torch.utils.data.Dataset` instance.")
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )

    def val_dataloader(self) -> DataLoader:
        """Returns the validation dataloader.

        Raises:
            ValueError: If validation dataset is not initialized.

        Returns:
            val dataloader.
        """
        if not self.val_dataset_available:
            raise ValueError("Validation dataset is not initialized")
        if not isinstance(self.val_dataset, torch.utils.data.Dataset):
            raise ValueError("Validation dataset has to be single `torch.utils.data.Dataset` instance.")
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )

    def test_dataloader(self) -> DataLoader:
        """Returns the test dataloader.

        Raises:
            ValueError: If test dataset is not initialized.


        Returns:
            test dataloader.
        """
        if not self.test_dataset_available:
            raise ValueError("Test dataset is not initialized")

        loader = DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )
        return loader

    def predict_dataloader(self) -> DataLoader:
        """Returns a dataloader used for predictions."""
        return self.test_dataloader()


class SklearnClassificationDataModule(BaseDataModule):
    """A generic Data Module for classification with frozen torch backbone and sklearn classifier.

    It can also handle k-fold cross validation.

    Args:
        name: The name for the data module. Defaults to "sklearn_classification_datamodule".
        data_path: Path to images main folder
        exclude_filter: List of string filter to be used to exclude images. If None no filter will be applied.
        include_filter: List of string filter to be used to include images. Only images that satisfied at list one of
                        the filter will be included.
        val_size: The validation split. Defaults to 0.2.
        class_to_idx: Dictionary of conversion btw folder name and index. Only file whose label is in dictionary key
            list will be considered. If None all files will be considered and a custom conversion is created.
        seed: Fixed seed for random operations
        batch_size: Dimension of batches for dataloader
        num_workers: Number of workers for dataloader
        train_transform: Albumentation transformations for training set
        val_transform: Albumentation transformations for validation set
        test_transform: Albumentation transformations for test set
        roi: Optional cropping region
        n_splits: Number of dataset subdivision (default 1 -> train/test). Use a value >= 2 for cross validation.
        phase: Either train or test
        cache: If true disable shuffling in all dataloader to enable feature caching
        limit_training_data: if defined, each class will be donwsampled to this number. It must be >= 2 to allow
            splitting
        label_map: Dictionary of conversion btw folder name and label.
        train_split_file: Optional path to a csv file containing the train split samples.
        test_split_file: Optional path to a csv file containing the test split samples.
        **kwargs: Additional arguments for BaseDataModule
    """

    def __init__(
        self,
        data_path: str,
        exclude_filter: list[str] | None = None,
        include_filter: list[str] | None = None,
        val_size: float = 0.2,
        class_to_idx: dict[str, int] | None = None,
        label_map: dict[str, Any] | None = None,
        seed: int = 42,
        batch_size: int = 32,
        num_workers: int = 6,
        train_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        roi: tuple[int, int, int, int] | None = None,
        n_splits: int = 1,
        phase: str = "train",
        cache: bool = False,
        limit_training_data: int | None = None,
        train_split_file: str | None = None,
        test_split_file: str | None = None,
        name: str = "sklearn_classification_datamodule",
        dataset: type[ImageClassificationListDataset] = ImageClassificationListDataset,
        **kwargs: Any,
    ):
        super().__init__(
            data_path=data_path,
            name=name,
            seed=seed,
            batch_size=batch_size,
            num_workers=num_workers,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            **kwargs,
        )

        self.class_to_idx = class_to_idx
        self.roi = roi
        self.cache = cache
        self.limit_training_data = limit_training_data

        self.dataset = dataset
        self.phase = phase
        self.n_splits = n_splits
        self.train_split_file = train_split_file
        self.test_split_file = test_split_file
        self.exclude_filter = exclude_filter
        self.include_filter = include_filter
        self.val_size = val_size
        self.label_map = label_map
        self.full_dataset: ImageClassificationListDataset
        self.train_dataset: list[ImageClassificationListDataset]
        self.val_dataset: list[ImageClassificationListDataset]

    def _prepare_data(self) -> None:
        """Prepares the data for the data module."""
        assert os.path.isdir(self.data_path), f"Folder {self.data_path} does not exist."

        list_df = []
        if self.phase == "train":
            samples, targets, split_generator, self.class_to_idx = get_split(
                image_dir=self.data_path,
                exclude_filter=self.exclude_filter,
                include_filter=self.include_filter,
                test_size=self.val_size,
                random_state=self.seed,
                class_to_idx=self.class_to_idx,
                n_splits=self.n_splits,
                limit_training_data=self.limit_training_data,
                train_split_file=self.train_split_file,
                label_map=self.label_map,
            )

            for cv_idx, split in enumerate(split_generator):
                train_idx, val_idx = split
                train_val_df = pd.DataFrame({"samples": samples, "targets": targets})
                train_val_df["cv"] = 0
                train_val_df["split"] = "train"
                train_val_df.loc[val_idx, "split"] = "val"
                train_val_df.loc[train_idx, "cv"] = cv_idx
                train_val_df.loc[val_idx, "cv"] = cv_idx
                list_df.append(train_val_df)

        test_samples, test_targets = find_test_image(
            folder=self.data_path,
            exclude_filter=self.exclude_filter,
            include_filter=self.include_filter,
            test_split_file=self.test_split_file,
        )
        if self.label_map is not None:
            test_targets, _ = group_labels(test_targets, self.label_map)
        test_df = pd.DataFrame({"samples": test_samples, "targets": test_targets})
        test_df["split"] = "test"
        test_df["cv"] = np.nan

        list_df.append(test_df)
        self.data = pd.concat(list_df, axis=0)

    def setup(self, stage: str) -> None:
        """Setup data module based on stages of training."""
        if stage == "fit":
            self.train_dataset = []
            self.val_dataset = []

            for cv_idx in range(self.n_splits):
                cv_df = self.data[self.data["cv"] == cv_idx]
                train_samples = cv_df[cv_df["split"] == "train"]["samples"].tolist()
                train_targets = cv_df[cv_df["split"] == "train"]["targets"].tolist()
                val_samples = cv_df[cv_df["split"] == "val"]["samples"].tolist()
                val_targets = cv_df[cv_df["split"] == "val"]["targets"].tolist()
                self.train_dataset.append(
                    self.dataset(
                        class_to_idx=self.class_to_idx,
                        samples=train_samples,
                        targets=train_targets,
                        transform=self.train_transform,
                        roi=self.roi,
                    )
                )
                self.val_dataset.append(
                    self.dataset(
                        class_to_idx=self.class_to_idx,
                        samples=val_samples,
                        targets=val_targets,
                        transform=self.val_transform,
                        roi=self.roi,
                    )
                )
            all_samples = self.data[self.data["cv"] == 0]["samples"].tolist()
            all_targets = self.data[self.data["cv"] == 0]["targets"].tolist()
            self.full_dataset = self.dataset(
                class_to_idx=self.class_to_idx,
                samples=all_samples,
                targets=all_targets,
                transform=self.train_transform,
                roi=self.roi,
            )
        if stage == "test":
            test_samples = self.data[self.data["split"] == "test"]["samples"].tolist()
            test_targets = self.data[self.data["split"] == "test"]["targets"]
            self.test_dataset = self.dataset(
                class_to_idx=self.class_to_idx,
                samples=test_samples,
                targets=test_targets.tolist(),
                transform=self.test_transform,
                roi=self.roi,
                allow_missing_label=True,
            )

    def predict_dataloader(self) -> DataLoader:
        """Returns a dataloader used for predictions."""
        return self.test_dataloader()

    def train_dataloader(self) -> list[DataLoader]:
        """Returns a list of train dataloader.

        Raises:
            ValueError: If train dataset is not initialized.

        Returns:
            list of train dataloader.
        """
        if not self.train_dataset_available:
            raise ValueError("Train dataset is not initialized")

        loader = []
        for dataset in self.train_dataset:
            loader.append(
                DataLoader(
                    dataset,
                    batch_size=self.batch_size,
                    shuffle=not self.cache,
                    num_workers=self.num_workers,
                    drop_last=False,
                    pin_memory=True,
                )
            )
        return loader

    def val_dataloader(self) -> list[DataLoader]:
        """Returns a list of validation dataloader.

        Raises:
            ValueError: If validation dataset is not initialized.

        Returns:
            List of validation dataloader.
        """
        if not self.val_dataset_available:
            raise ValueError("Validation dataset is not initialized")

        loader = []
        for dataset in self.val_dataset:
            loader.append(
                DataLoader(
                    dataset,
                    batch_size=self.batch_size,
                    shuffle=False,
                    num_workers=self.num_workers,
                    drop_last=False,
                    pin_memory=True,
                )
            )

        return loader

    def test_dataloader(self) -> DataLoader:
        """Returns the test dataloader.

        Raises:
            ValueError: If test dataset is not initialized.


        Returns:
            test dataloader.
        """
        if not self.test_dataset_available:
            raise ValueError("Test dataset is not initialized")

        loader = DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )
        return loader

    def full_dataloader(self) -> DataLoader:
        """Return a dataloader to perform training on the entire dataset.

        Returns:
            dataloader to perform training on the entire dataset after evaluation. This is useful
            to perform a final training on the entire dataset after the evaluation phase.

        """
        if self.full_dataset is None:
            raise ValueError("Full dataset is not initialized")

        return DataLoader(
            self.full_dataset,
            batch_size=self.batch_size,
            shuffle=not self.cache,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
        )


class MultilabelClassificationDataModule(BaseDataModule):
    """Base class for all multi-label modules.

    Args:
        data_path: Path to the data main folder.
        images_and_labels_file: a path to a txt file containing the relative (to `data_path`) path
            of images with their relative labels, in a comma-separated way.
            E.g.:

             * path1,l1,l2,l3
             * path2,l4,l5
             * ...

            One of `images_and_label` and both `train_split_file` and `test_split_file` must be set.
            Defaults to None.
        name: The name for the data module. Defaults to "multilabel_datamodule".
        dataset: a callable returning a torch.utils.data.Dataset class.
        num_classes: the number of classes in the dataset. This is used to create one-hot encoded
            targets. Defaults to None.
        num_workers: Number of workers for dataloaders. Defaults to 16.
        batch_size: Training batch size. Defaults to 64.
        test_batch_size: Testing batch size. Defaults to 64.
        seed: Random generator seed. Defaults to SegmentationEvalua2.
        val_size: The validation split. Defaults to 0.2.
        test_size: The test split. Defaults to 0.2.
        train_transform: Transformations for train dataset.
            Defaults to None.
        val_transform: Transformations for validation dataset.
            Defaults to None.
        test_transform: Transformations for test dataset.
            Defaults to None.
        train_split_file: The file with train split. Defaults to None.
        val_split_file: The file with validation split. Defaults to None.
        test_split_file: The file with test split. Defaults to None.
        class_to_idx: a clss to idx dictionary. Defaults to None.
    """

    def __init__(
        self,
        data_path: str,
        images_and_labels_file: str | None = None,
        train_split_file: str | None = None,
        test_split_file: str | None = None,
        val_split_file: str | None = None,
        name: str = "multilabel_datamodule",
        dataset: Callable = MultilabelClassificationDataset,
        num_classes: int | None = None,
        num_workers: int = 16,
        batch_size: int = 64,
        test_batch_size: int = 64,
        seed: int = 42,
        val_size: float | None = 0.2,
        test_size: float | None = 0.2,
        train_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        class_to_idx: dict[str, int] | None = None,
        **kwargs,
    ):
        super().__init__(
            data_path=data_path,
            name=name,
            num_workers=num_workers,
            batch_size=batch_size,
            seed=seed,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            **kwargs,
        )
        if not (images_and_labels_file is not None or (train_split_file is not None and test_split_file is not None)):
            raise ValueError(
                "Either `images_and_labels_file` or both `train_split_file` and `test_split_file` must be set"
            )
        self.images_and_labels_file = images_and_labels_file
        self.dataset = dataset
        self.num_classes = num_classes
        self.train_batch_size = batch_size
        self.test_batch_size = test_batch_size
        self.val_size = val_size
        self.test_size = test_size
        self.train_split_file = train_split_file
        self.test_split_file = test_split_file
        self.val_split_file = val_split_file
        self.class_to_idx = class_to_idx
        self.train_dataset: MultilabelClassificationDataset
        self.val_dataset: MultilabelClassificationDataset
        self.test_dataset: MultilabelClassificationDataset

    def _read_split(self, split_file: str) -> tuple[list[str], list[list[str]]]:
        """Reads split file.

        Args:
            split_file: Path to the split file.

        Returns:
            Tuple containing list of paths to images and list of labels.
        """
        all_samples, all_targets = [], []
        with open(split_file) as f:
            for line in f.readlines():
                split_line = line.split(",")
                sample = os.path.join(self.data_path, split_line[0])
                targets = [t.strip() for t in split_line[1:]]
                if len(targets) == 0:
                    continue
                all_samples.append(sample)
                all_targets.append(targets)
        return all_samples, all_targets

    def _prepare_data(self) -> None:
        """Prepares the data for the data module."""
        if self.images_and_labels_file is not None:
            # Read all images and targets
            all_samples, all_targets = self._read_split(self.images_and_labels_file)
            all_samples = np.array(all_samples).reshape(-1, 1)

            # Targets to idx
            unique_targets = set(utils.flatten_list(all_targets))
            if self.class_to_idx is None:
                self.class_to_idx = {c: i for i, c in enumerate(unique_targets)}

            all_targets = [[self.class_to_idx[t] for t in targets] for targets in all_targets]

            # Transform targets to one-hot
            if self.num_classes is None:
                self.num_classes = len(unique_targets)
            all_targets = np.array([[i in targets for i in range(self.num_classes)] for targets in all_targets]).astype(
                int
            )

            # Create splits
            samples_train, targets_train, samples_test, targets_test = iterative_train_test_split(
                all_samples, all_targets, test_size=self.test_size
            )
        elif self.train_split_file is not None and self.test_split_file is not None:
            # Both train_split_file and test_split_file are set
            samples_train, targets_train = self._read_split(self.train_split_file)
            samples_test, targets_test = self._read_split(self.test_split_file)

            # Create class_to_idx from all targets
            unique_targets = set(utils.flatten_list(targets_test + targets_train))
            if self.class_to_idx is None:
                self.class_to_idx = {c: i for i, c in enumerate(unique_targets)}

            # Transform targets to one-hot
            if self.num_classes is None:
                self.num_classes = len(unique_targets)
            targets_test = [[self.class_to_idx[t] for t in targets] for targets in targets_test]
            targets_test = np.array(
                [[i in targets for i in range(self.num_classes)] for targets in targets_test]
            ).astype(int)
            targets_train = [[self.class_to_idx[t] for t in targets] for targets in targets_train]
            targets_train = np.array(
                [[i in targets for i in range(self.num_classes)] for targets in targets_train]
            ).astype(int)
        else:
            raise ValueError(
                "Either `images_and_labels_file` or both `train_split_file` and `test_split_file` must be set"
            )

        if self.val_split_file:
            if not self.test_split_file or not self.train_split_file:
                raise ValueError("Validation split file is specified but no train or test split file is specified.")
            samples_val, targets_val = self._read_split(self.val_split_file)
            targets_val = [[self.class_to_idx[t] for t in targets] for targets in targets_val]
            targets_val = np.array([[i in targets for i in range(self.num_classes)] for targets in targets_val]).astype(
                int
            )
        else:
            samples_train = np.array(samples_train).reshape(-1, 1)
            targets_train = np.array(targets_train).reshape(-1, self.num_classes)
            samples_train, targets_train, samples_val, targets_val = iterative_train_test_split(
                samples_train, targets_train, test_size=self.val_size
            )

        if isinstance(samples_train, np.ndarray):
            samples_train = samples_train.flatten().tolist()
        if isinstance(samples_val, np.ndarray):
            samples_val = samples_val.flatten().tolist()
        if isinstance(samples_test, np.ndarray):
            samples_test = samples_test.flatten().tolist()

        if isinstance(targets_train, np.ndarray):
            targets_train = list(targets_train)
        if isinstance(targets_val, np.ndarray):
            targets_val = list(targets_val)  # type: ignore[assignment]
        if isinstance(targets_test, np.ndarray):
            targets_test = list(targets_test)

        # Create data
        train_df = pd.DataFrame({"samples": samples_train, "targets": targets_train})
        train_df["split"] = "train"
        val_df = pd.DataFrame({"samples": samples_val, "targets": targets_val})
        val_df["split"] = "val"
        test_df = pd.DataFrame({"samples": samples_test, "targets": targets_test})
        test_df["split"] = "test"
        self.data = pd.concat([train_df, val_df, test_df], axis=0)

    def setup(self, stage: str | None = None) -> None:
        """Setup data module based on stages of training."""
        if stage in ["train", "fit"]:
            train_samples = self.data[self.data["split"] == "train"]["samples"].tolist()
            train_targets = self.data[self.data["split"] == "train"]["targets"].tolist()
            val_samples = self.data[self.data["split"] == "val"]["samples"].tolist()
            val_targets = self.data[self.data["split"] == "val"]["targets"].tolist()
            self.train_dataset = self.dataset(
                samples=train_samples,
                targets=train_targets,
                transform=self.train_transform,
                class_to_idx=self.class_to_idx,
            )
            self.val_dataset = self.dataset(
                samples=val_samples,
                targets=val_targets,
                transform=self.val_transform,
                class_to_idx=self.class_to_idx,
            )
        if stage == "test":
            test_samples = self.data[self.data["split"] == "test"]["samples"].tolist()
            test_targets = self.data[self.data["split"] == "test"]["targets"].tolist()
            self.test_dataset = self.dataset(
                samples=test_samples,
                targets=test_targets,
                transform=self.test_transform,
                class_to_idx=self.class_to_idx,
            )

    def train_dataloader(self) -> DataLoader:
        """Returns the train dataloader.

        Raises:
            ValueError: If train dataset is not initialized.

        Returns:
            Train dataloader.
        """
        if not self.train_dataset_available:
            raise ValueError("Train dataset is not initialized")
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )

    def val_dataloader(self) -> DataLoader:
        """Returns the validation dataloader.

        Raises:
            ValueError: If validation dataset is not initialized.

        Returns:
            val dataloader.
        """
        if not self.val_dataset_available:
            raise ValueError("Validation dataset is not initialized")
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )

    def test_dataloader(self) -> DataLoader:
        """Returns the test dataloader.

        Raises:
            ValueError: If test dataset is not initialized.


        Returns:
            test dataloader.
        """
        if not self.test_dataset_available:
            raise ValueError("Test dataset is not initialized")

        loader = DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=False,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
        )
        return loader

    def predict_dataloader(self) -> DataLoader:
        """Returns a dataloader used for predictions."""
        return self.test_dataloader()



import hydra
import pytorch_lightning as pl
from omegaconf import DictConfig
from pytorch_lightning import Callback
from pytorch_lightning.utilities import rank_zero_only

from quadra.schedulers.warmup import CosineAnnealingWithLinearWarmUp
from quadra.utils.utils import get_logger

log = get_logger(__name__)


class WarmupInit(Callback):
    """Custom callback used to setup a warmup scheduler.

    Args:
        scheduler_config: scheduler configuration.
    """

    def __init__(
        self,
        scheduler_config: DictConfig,
    ) -> None:
        self.scheduler_config = scheduler_config

    @rank_zero_only
    def on_fit_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Called when fit begins."""
        if not hasattr(trainer, "datamodule"):
            raise ValueError("Trainer must have a datamodule attribute.")

        if not any(isinstance(s.scheduler, CosineAnnealingWithLinearWarmUp) for s in trainer.lr_scheduler_configs):
            return

        log.info("Using warmup scheduler, forcing optimizer learning rate to zero.")
        for i, _ in enumerate(trainer.optimizers):
            for param_group in trainer.optimizers[i].param_groups:
                param_group["lr"] = 0.0
            trainer.optimizers[i].defaults["lr"] = 0.0

        batch_size = trainer.datamodule.batch_size
        train_dataloader = trainer.datamodule.train_dataloader()
        len_train_dataloader = len(train_dataloader)
        if isinstance(trainer.device_ids, list) and pl_module.device.type == "cuda":
            num_gpus = len(trainer.device_ids)
            len_train_dataloader = len_train_dataloader // num_gpus
            if not train_dataloader.drop_last:
                len_train_dataloader += int((len_train_dataloader % num_gpus) != 0)

        if len_train_dataloader == 1:
            log.warning(
                "From this dataset size, we can only generate single batch. The batch size will be set as lenght of"
                " the dataset "
            )
            batch_size = len(train_dataloader.dataset)

        if isinstance(trainer.device_ids, list) and pl_module.device.type == "cuda":
            batch_size = batch_size * len(trainer.device_ids)

        scheduler = hydra.utils.instantiate(
            self.scheduler_config,
            optimizer=pl_module.optimizer,
            batch_size=batch_size,
            len_loader=len_train_dataloader,
        )

        for i, s in enumerate(trainer.lr_scheduler_configs):
            if isinstance(s.scheduler, CosineAnnealingWithLinearWarmUp):
                trainer.lr_scheduler_configs[i].scheduler = scheduler



import logging

from pytorch_lightning.utilities import rank_zero_only


def get_logger(name=__name__) -> logging.Logger:
    """Initializes multi-GPU-friendly python logger."""
    logger = logging.getLogger(name)

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    for level in ("debug", "info", "warning", "error", "exception", "fatal", "critical"):
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger



import torch
import torch.nn.functional as F


def idmm_loss(
    p1: torch.Tensor,
    y1: torch.Tensor,
    smoothing: float = 0.1,
) -> torch.Tensor:
    """IDMM loss described in https://arxiv.org/abs/2201.10728.

    Args:
        p1: Prediction labels for `z1`
        y1: Instance labels for `z1`
        smoothing: smoothing factor used for label smoothing.
            Defaults to 0.1.

    Returns:
        IDMM loss
    """
    loss = F.cross_entropy(p1, y1, label_smoothing=smoothing)
    return loss


class IDMMLoss(torch.nn.Module):
    """IDMM loss described in https://arxiv.org/abs/2201.10728."""

    def __init__(self, smoothing: float = 0.1):
        super().__init__()
        self.smoothing = smoothing

    def forward(
        self,
        p1: torch.Tensor,
        y1: torch.Tensor,
    ) -> torch.Tensor:
        """IDMM loss described in https://arxiv.org/abs/2201.10728.

        Args:
            p1: Prediction labels for `z1`
            y1: Instance labels for `z1`

        Returns:
            IDMM loss
        """
        return idmm_loss(
            p1,
            y1,
            self.smoothing,
        )



from __future__ import annotations

import sklearn
import torch
from torch import nn, optim

from quadra.modules.base import SSLModule


class VICReg(SSLModule):
    """VICReg model.

    Args:
        model: Network Module used for extract features
        projection_mlp: Module to project extracted features
        criterion: SSL loss to be applied.
        classifier: Standard sklearn classifier. Defaults to None.
        optimizer: optimizer of the training. If None a default Adam is used. Defaults to None.
        lr_scheduler: lr scheduler. If None a default ReduceLROnPlateau is used. Defaults to None.
        lr_scheduler_interval: interval at which the lr scheduler is updated. Defaults to "epoch".

    """

    def __init__(
        self,
        model: nn.Module,
        projection_mlp: nn.Module,
        criterion: nn.Module,
        classifier: sklearn.base.ClassifierMixin | None = None,
        optimizer: optim.Optimizer | None = None,
        lr_scheduler: object | None = None,
        lr_scheduler_interval: str | None = "epoch",
    ):
        super().__init__(
            model,
            criterion,
            classifier,
            optimizer,
            lr_scheduler,
            lr_scheduler_interval,
        )
        # self.save_hyperparameters()
        self.projection_mlp = projection_mlp
        self.criterion = criterion

    def forward(self, x):
        x = self.model(x)
        z = self.projection_mlp(x)
        return z

    def training_step(self, batch: tuple[tuple[torch.Tensor, torch.Tensor], torch.Tensor], batch_idx: int):
        # pylint: disable=unused-argument
        # Compute loss
        (im_x, im_y), _ = batch
        z1 = self(im_x)
        z2 = self(im_y)
        loss = self.criterion(z1, z2)

        self.log(
            "loss",
            loss,
            on_epoch=True,
            on_step=True,
            logger=True,
            prog_bar=True,
        )
        return loss



from __future__ import annotations

import os
import pathlib

import albumentations
import pandas as pd
from torch.utils.data import DataLoader

from quadra.datamodules.base import BaseDataModule
from quadra.datasets import AnomalyDataset
from quadra.datasets.anomaly import make_anomaly_dataset
from quadra.utils import utils

log = utils.get_logger(__name__)


class AnomalyDataModule(BaseDataModule):
    """Anomalib-like Lightning Data Module.

    Args:
        data_path: Path to the dataset
        category: Name of the sub category to use.
        image_size: Variable to which image is resized.
        train_batch_size: Training batch size.
        test_batch_size: Testing batch size.
        train_transform: transformations for training. Defaults to None.
        val_transform: transformations for validation. Defaults to None.
        test_transform: transformations for testing. Defaults to None.
        num_workers: Number of workers.
        seed: seed used for the random subset splitting
        task: Whether we are interested in segmenting the anomalies (segmentation) or not (classification)
        mask_suffix: String to append to the base filename to get the mask name, by default for MVTec dataset masks
            are saved as imagename_mask.png in this case the parameter should be filled with "_mask"
        create_test_set_if_empty: If True, the test set is created from good images if it is empty.
        phase: Either train or test.
        name: Name of the data module.
        valid_area_mask: Optional path to the mask to use to filter out the valid area of the image. If None, the whole
            image is considered valid. The mask should match the image size even if the image is cropped.
        crop_area: Optional tuple of 4 integers (x1, y1, x2, y2) to crop the image to the specified area. If None, the
            whole image is considered valid.
    """

    def __init__(
        self,
        data_path: str,
        category: str | None = None,
        image_size: int | tuple[int, int] | None = None,
        train_batch_size: int = 32,
        test_batch_size: int = 32,
        num_workers: int = 8,
        train_transform: albumentations.Compose | None = None,
        val_transform: albumentations.Compose | None = None,
        test_transform: albumentations.Compose | None = None,
        seed: int = 0,
        task: str = "segmentation",
        mask_suffix: str | None = None,
        create_test_set_if_empty: bool = True,
        phase: str = "train",
        name: str = "anomaly_datamodule",
        valid_area_mask: str | None = None,
        crop_area: tuple[int, int, int, int] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            data_path=data_path,
            name=name,
            seed=seed,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            num_workers=num_workers,
            **kwargs,
        )

        self.root = data_path
        self.category = category
        self.data_path = os.path.join(self.root, self.category) if self.category is not None else self.root
        self.image_size = image_size

        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        self.task = task

        self.train_dataset: AnomalyDataset
        self.test_dataset: AnomalyDataset
        self.val_dataset: AnomalyDataset
        self.mask_suffix = mask_suffix
        self.create_test_set_if_empty = create_test_set_if_empty
        self.phase = phase
        self.valid_area_mask = valid_area_mask
        self.crop_area = crop_area

    @property
    def val_data(self) -> pd.DataFrame:
        """Get validation data."""
        _val_data = super().val_data
        if len(_val_data) == 0:
            return self.test_data
        return _val_data

    def _prepare_data(self) -> None:
        """Prepare data for training and testing."""
        self.data = make_anomaly_dataset(
            path=pathlib.Path(self.data_path),
            split=None,
            seed=self.seed,
            mask_suffix=self.mask_suffix,
            create_test_set_if_empty=self.create_test_set_if_empty,
        )

    def setup(self, stage: str | None = None) -> None:
        """Setup data module based on stages of training."""
        if stage == "fit" and self.phase == "train":
            self.train_dataset = AnomalyDataset(
                transform=self.train_transform,
                task=self.task,
                samples=self.train_data,
                valid_area_mask=self.valid_area_mask,
                crop_area=self.crop_area,
            )

            if len(self.val_data) == 0:
                log.info("Validation dataset is empty, using test set instead")

            self.val_dataset = AnomalyDataset(
                transform=self.test_transform,
                task=self.task,
                samples=self.val_data if len(self.val_data) > 0 else self.data,
                valid_area_mask=self.valid_area_mask,
                crop_area=self.crop_area,
            )
        if stage == "test" or self.phase == "test":
            self.test_dataset = AnomalyDataset(
                transform=self.test_transform,
                task=self.task,
                samples=self.test_data,
                valid_area_mask=self.valid_area_mask,
                crop_area=self.crop_area,
            )

    def train_dataloader(self) -> DataLoader:
        """Get train dataloader."""
        return DataLoader(
            self.train_dataset,
            shuffle=True,
            batch_size=self.train_batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def val_dataloader(self) -> DataLoader:
        """Get validation dataloader."""
        return DataLoader(
            dataset=self.val_dataset,
            shuffle=False,
            batch_size=self.test_batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def test_dataloader(self) -> DataLoader:
        """Get test dataloader."""
        return DataLoader(
            self.test_dataset,
            shuffle=False,
            batch_size=self.test_batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def predict_dataloader(self) -> DataLoader:
        """Returns a dataloader used for predictions."""
        return DataLoader(
            self.test_dataset,
            shuffle=False,
            batch_size=self.test_batch_size,
            num_workers=self.num_workers,
            pin_memory=True,
        )



import functools
from collections.abc import Callable

from quadra.utils.utils import get_logger

logger = get_logger(__name__)


def deprecated(message: str) -> Callable:
    """Decorator to mark a function as deprecated.

    Args:
        message: Message to be displayed when the function is called.

    Returns:
        Decoratored function.
    """

    def deprecated_decorator(func_or_class: Callable) -> Callable:
        """Decorator to mark a function as deprecated."""

        @functools.wraps(func_or_class)
        def wrapper(*args, **kwargs):
            """Wrapper function to display a warning message."""
            warning_msg = f"{func_or_class.__name__} is deprecated. {message}"
            logger.warning(warning_msg)
            return func_or_class(*args, **kwargs)

        return wrapper

    return deprecated_decorator



from .anomaly import AnomalyDatasetArguments, anomaly_dataset, base_anomaly_dataset
from .classification import (
    ClassificationDatasetArguments,
    ClassificationMultilabelDatasetArguments,
    ClassificationPatchDatasetArguments,
    base_classification_dataset,
    base_multilabel_classification_dataset,
    base_patch_classification_dataset,
    classification_dataset,
    classification_patch_dataset,
    multilabel_classification_dataset,
)
from .imagenette import imagenette_dataset
from .segmentation import (
    SegmentationDatasetArguments,
    base_binary_segmentation_dataset,
    base_multiclass_segmentation_dataset,
    segmentation_dataset,
)

__all__ = [
    "anomaly_dataset",
    "classification_dataset",
    "AnomalyDatasetArguments",
    "ClassificationDatasetArguments",
    "ClassificationPatchDatasetArguments",
    "classification_patch_dataset",
    "segmentation_dataset",
    "SegmentationDatasetArguments",
    "multilabel_classification_dataset",
    "ClassificationMultilabelDatasetArguments",
    "base_anomaly_dataset",
    "imagenette_dataset",
    "base_classification_dataset",
    "base_patch_classification_dataset",
    "base_binary_segmentation_dataset",
    "base_multiclass_segmentation_dataset",
    "base_multilabel_classification_dataset",
]



from __future__ import annotations

import os
from pathlib import Path

import numpy as np
import torch
from hydra import compose, initialize_config_module
from hydra.core.hydra_config import HydraConfig

from quadra.main import main
from quadra.utils.export import get_export_extension


# taken from hydra unit tests
def _random_image(size: tuple[int, int] = (10, 10)) -> np.ndarray:
    """Generate random image."""
    return np.random.randint(0, 255, size=size, dtype=np.uint8)


def execute_quadra_experiment(overrides: list[str], experiment_path: Path) -> None:
    """Execute quadra experiment."""
    with initialize_config_module(config_module="quadra.configs", version_base="1.3.0"):
        if not experiment_path.exists():
            experiment_path.mkdir(parents=True)
        os.chdir(experiment_path)
        # cfg = compose(config_name="config", overrides=overrides)
        cfg = compose(config_name="config", overrides=overrides, return_hydra_config=True)
        # workaround without actual main function
        # check https://github.com/facebookresearch/hydra/issues/2017 for more details
        HydraConfig.instance().set_config(cfg)

        main(cfg)


def check_deployment_model(export_type: str):
    """Check that the runtime model is present and valid.

    Args:
        export_type: The type of the exported model.
    """
    extension = get_export_extension(export_type)

    assert os.path.exists(f"deployment_model/model.{extension}")
    assert os.path.exists("deployment_model/model.json")


def get_quadra_test_device():
    """Get the device to use for the tests. If the QUADRA_TEST_DEVICE environment variable is set, it is used."""
    return os.environ.get("QUADRA_TEST_DEVICE", "cpu")


def setup_trainer_for_lightning() -> list[str]:
    """Setup trainer for lightning depending on the device. If cuda is used, the device index is also set.
    If cpu is used, the trainer is set to lightning_cpu.

    Returns:
        A list of overrides for the trainer.
    """
    overrides = []
    device = get_quadra_test_device()
    torch_device = torch.device(device)
    if torch_device.type == "cuda":
        device_index = torch_device.index
        overrides.append("trainer=lightning_gpu")
        overrides.append(f"trainer.devices=[{device_index}]")
    else:
        overrides.append("trainer=lightning_cpu")

    return overrides



from __future__ import annotations

import json
import os
import typing
from typing import Any, Generic

import cv2
import hydra
import torch
from omegaconf import DictConfig, OmegaConf
from torch.utils.data import DataLoader

from quadra.callbacks.mlflow import get_mlflow_logger
from quadra.datamodules import SegmentationDataModule, SegmentationMulticlassDataModule
from quadra.models.base import ModelSignatureWrapper
from quadra.models.evaluation import BaseEvaluationModel
from quadra.modules.base import SegmentationModel
from quadra.tasks.base import Evaluation, LightningTask
from quadra.utils import utils
from quadra.utils.evaluation import automatic_datamodule_batch_size, create_mask_report
from quadra.utils.export import export_model

log = utils.get_logger(__name__)

SegmentationDataModuleT = typing.TypeVar(
    "SegmentationDataModuleT", SegmentationDataModule, SegmentationMulticlassDataModule
)


class Segmentation(Generic[SegmentationDataModuleT], LightningTask[SegmentationDataModuleT]):
    """Task for segmentation.

    Args:
        config: Config object
        num_viz_samples: Number of samples to visualize. Defaults to 5.
        checkpoint_path: Path to the checkpoint to load the model from. Defaults to None.
        run_test: If True, run test after training. Defaults to False.
        evaluate: Dict with evaluation parameters. Defaults to None.
        report: If True, create report after training. Defaults to False.
    """

    def __init__(
        self,
        config: DictConfig,
        num_viz_samples: int = 5,
        checkpoint_path: str | None = None,
        run_test: bool = False,
        evaluate: DictConfig | None = None,
        report: bool = False,
    ):
        super().__init__(
            config=config,
            checkpoint_path=checkpoint_path,
            run_test=run_test,
            report=report,
        )
        self.evaluate = evaluate
        self.num_viz_samples = num_viz_samples
        self.export_folder: str = "deployment_model"
        self.exported_model_path: str | None = None
        if self.evaluate and any(self.evaluate.values()):
            if (
                self.config.export is None
                or len(self.config.export.types) == 0
                or "torchscript" not in self.config.export.types
            ):
                log.info(
                    "Evaluation is enabled, but training does not export a deployment model. Automatically export the "
                    "model as torchscript."
                )
                if self.config.export is None:
                    self.config.export = DictConfig({"types": ["torchscript"]})
                else:
                    self.config.export.types.append("torchscript")

            if not self.report:
                log.info("Evaluation is enabled, but reporting is disabled. Enabling reporting automatically.")
                self.report = True

    @property
    def module(self) -> SegmentationModel:
        """Get the module."""
        return self._module

    @module.setter
    def module(self, module_config) -> None:
        """Set the module."""
        log.info("Instantiating model <%s>", module_config.model["_target_"])

        if isinstance(self.datamodule, SegmentationMulticlassDataModule) and module_config.model.num_classes != (
            len(self.datamodule.idx_to_class) + 1
        ):
            log.warning(
                "Number of classes in the model (%s) does not match the number of "
                + "classes in the datamodule (%d). Updating the model...",
                module_config.model.num_classes,
                len(self.datamodule.idx_to_class),
            )
            module_config.model.num_classes = len(self.datamodule.idx_to_class) + 1

        model = hydra.utils.instantiate(module_config.model)
        model = ModelSignatureWrapper(model)
        log.info("Instantiating optimizer <%s>", self.config.optimizer["_target_"])
        param_list = []
        for param in model.parameters():
            if param.requires_grad:
                param_list.append(param)
        optimizer = hydra.utils.instantiate(self.config.optimizer, param_list)
        log.info("Instantiating scheduler <%s>", self.config.scheduler["_target_"])
        scheduler = hydra.utils.instantiate(self.config.scheduler, optimizer=optimizer)
        log.info("Instantiating module <%s>", module_config.module["_target_"])
        module = hydra.utils.instantiate(module_config.module, model=model, optimizer=optimizer, lr_scheduler=scheduler)
        if self.checkpoint_path is not None:
            module.__class__.load_from_checkpoint(
                self.checkpoint_path, model=model, optimizer=optimizer, lr_scheduler=scheduler
            )
        self._module = module

    def prepare(self) -> None:
        """Prepare the task."""
        super().prepare()
        self.module = self.config.model

    def export(self) -> None:
        """Generate a deployment model for the task."""
        log.info("Exporting model ready for deployment")

        # Get best model!
        if (
            self.trainer.checkpoint_callback is not None
            and hasattr(self.trainer.checkpoint_callback, "best_model_path")
            and self.trainer.checkpoint_callback.best_model_path is not None
            and len(self.trainer.checkpoint_callback.best_model_path) > 0
        ):
            best_model_path = self.trainer.checkpoint_callback.best_model_path
            log.info("Loaded best model from %s", best_model_path)

            module = self.module.__class__.load_from_checkpoint(
                best_model_path,
                model=self.module.model,
                loss_fun=None,
                optimizer=self.module.optimizer,
                lr_scheduler=self.module.schedulers,
            )
        else:
            log.warning("No checkpoint callback found in the trainer, exporting the last model weights")
            module = self.module

        if "idx_to_class" not in self.config.datamodule:
            log.info("No idx_to_class key")
            idx_to_class = {0: "good", 1: "bad"}  # TODO: Why is this the default value?
        else:
            log.info("idx_to_class is present")
            idx_to_class = self.config.datamodule.idx_to_class

        if self.config.export is None:
            raise ValueError(
                "No export type specified. This should not happen, please check if you have set "
                "the export_type or assign it to a default value."
            )

        half_precision = "16" in self.trainer.precision

        input_shapes = self.config.export.input_shapes

        model_json, export_paths = export_model(
            config=self.config,
            model=module.model,
            export_folder=self.export_folder,
            half_precision=half_precision,
            input_shapes=input_shapes,
            idx_to_class=idx_to_class,
        )

        if len(export_paths) == 0:
            return

        # Pick one model for evaluation, it should be independent of the export type as the model is wrapped
        self.exported_model_path = next(iter(export_paths.values()))

        with open(os.path.join(self.export_folder, "model.json"), "w") as f:
            json.dump(model_json, f, cls=utils.HydraEncoder)

    def generate_report(self) -> None:
        """Generate a report for the task."""
        if self.evaluate is not None:
            log.info("Generating evaluation report!")
            eval_tasks: list[SegmentationEvaluation] = []
            if self.evaluate.analysis:
                if self.exported_model_path is None:
                    raise ValueError(
                        "Exported model path is not set yet but the task tries to do an analysis evaluation"
                    )
                eval_task = SegmentationAnalysisEvaluation(
                    config=self.config,
                    model_path=self.exported_model_path,
                )
                eval_tasks.append(eval_task)
            for task in eval_tasks:
                task.execute()

            if len(self.logger) > 0:
                mflow_logger = get_mlflow_logger(trainer=self.trainer)
                tensorboard_logger = utils.get_tensorboard_logger(trainer=self.trainer)

                if mflow_logger is not None and self.config.core.get("upload_artifacts"):
                    for task in eval_tasks:
                        for file in task.metadata["report_files"]:
                            mflow_logger.experiment.log_artifact(
                                run_id=mflow_logger.run_id, local_path=file, artifact_path=task.report_path
                            )

                if tensorboard_logger is not None and self.config.core.get("upload_artifacts"):
                    for task in eval_tasks:
                        for file in task.metadata["report_files"]:
                            ext = os.path.splitext(file)[1].lower()

                            if ext in [".png", ".jpg", ".jpeg", ".bmp", ".tiff", ".tif", ".gif"]:
                                try:
                                    img = cv2.imread(file)
                                    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                                except cv2.error:
                                    log.info("Could not upload artifact image %s", file)
                                    continue

                                tensorboard_logger.experiment.add_image(
                                    os.path.basename(file), img, 0, dataformats="HWC"
                                )
                            else:
                                utils.upload_file_tensorboard(file, tensorboard_logger)


class SegmentationEvaluation(Evaluation[SegmentationDataModuleT]):
    """Segmentation Evaluation Task with deployment models.

    Args:
        config: The experiment configuration
        model_path: The experiment path.
        device: Device to use for evaluation. If None, the device is automatically determined.

    Raises:
        ValueError: If the model path is not provided
    """

    def __init__(
        self,
        config: DictConfig,
        model_path: str,
        device: str | None = "cpu",
    ):
        super().__init__(config=config, model_path=model_path, device=device)
        self.config = config

    def save_config(self) -> None:
        """Skip saving the config."""

    def prepare(self) -> None:
        """Prepare the evaluation."""
        super().prepare()
        # TODO: Why we propagate mean and std only in Segmentation?
        self.config.transforms.mean = self.model_data["mean"]
        self.config.transforms.std = self.model_data["std"]
        # Setup datamodule
        if hasattr(self.config.datamodule, "idx_to_class"):
            idx_to_class = self.model_data["classes"]  # dict {index: class}
            self.config.datamodule.idx_to_class = idx_to_class
        self.datamodule = self.config.datamodule
        # prepare_data() must be explicitly called because there is no lightning training
        self.datamodule.prepare_data()

    @torch.no_grad()
    def inference(
        self, dataloader: DataLoader, deployment_model: BaseEvaluationModel, device: torch.device
    ) -> dict[str, torch.Tensor]:
        """Run inference on the dataloader and return the output.

        Args:
            dataloader: The dataloader to run inference on
            deployment_model: The deployment model to use
            device: The device to run inference on
        """
        image_list, mask_list, mask_pred_list, label_list = [], [], [], []
        for batch in dataloader:
            images, masks, labels = batch
            images = images.to(device)
            masks = masks.to(device)
            labels = labels.to(device)
            image_list.append(images.cpu())
            mask_list.append(masks.cpu())
            mask_pred_list.append(deployment_model(images.to(device)).cpu())
            label_list.append(labels.cpu())
        output = {
            "image": torch.cat(image_list, dim=0),
            "mask": torch.cat(mask_list, dim=0),
            "label": torch.cat(label_list, dim=0),
            "mask_pred": torch.cat(mask_pred_list, dim=0),
        }
        return output


class SegmentationAnalysisEvaluation(SegmentationEvaluation):
    """Segmentation Analysis Evaluation Task
    Args:
        config: The experiment configuration
        model_path: The model path.
        device: Device to use for evaluation. If None, the device is automatically determined.
    """

    def __init__(
        self,
        config: DictConfig,
        model_path: str,
        device: str | None = None,
    ):
        super().__init__(config=config, model_path=model_path, device=device)
        self.test_output: dict[str, Any] = {}

    def train(self) -> None:
        """Skip training."""

    def prepare(self) -> None:
        """Prepare the evaluation task."""
        super().prepare()
        self.datamodule.setup(stage="fit")
        self.datamodule.setup(stage="test")

    @automatic_datamodule_batch_size(batch_size_attribute_name="batch_size")
    def test(self) -> None:
        """Run testing."""
        log.info("Starting inference for analysis.")

        stages: list[str] = []
        dataloaders: list[torch.utils.data.DataLoader] = []

        # if self.datamodule.train_dataset_available:
        #     stages.append("train")
        #     dataloaders.append(self.datamodule.train_dataloader())
        #     if self.datamodule.val_dataset_available:
        #         stages.append("val")
        #         dataloaders.append(self.datamodule.val_dataloader())

        if self.datamodule.test_dataset_available:
            stages.append("test")
            dataloaders.append(self.datamodule.test_dataloader())
        for stage, dataloader in zip(stages, dataloaders, strict=False):
            log.info("Running inference on %s set with batch size: %d", stage, dataloader.batch_size)
            image_list, mask_list, mask_pred_list, label_list = [], [], [], []
            for batch in dataloader:
                images, masks, labels = batch
                images = images.to(device=self.device, dtype=self.deployment_model.model_dtype)
                if len(masks.shape) == 3:  # BxHxW -> Bx1xHxW
                    masks = masks.unsqueeze(1)
                with torch.no_grad():
                    image_list.append(images)
                    mask_list.append(masks)
                    mask_pred_list.append(self.deployment_model(images.to(self.device)))
                    label_list.append(labels)

            output = {
                "image": torch.cat(image_list, dim=0),
                "mask": torch.cat(mask_list, dim=0),
                "label": torch.cat(label_list, dim=0),
                "mask_pred": torch.cat(mask_pred_list, dim=0),
            }
            self.test_output[stage] = output

    def generate_report(self) -> None:
        """Generate a report."""
        log.info("Generating analysis report")

        for stage, output in self.test_output.items():
            image_mean = OmegaConf.to_container(self.config.transforms.mean)
            if not isinstance(image_mean, list) or any(not isinstance(x, int | float) for x in image_mean):
                raise ValueError("Image mean is not a list of float or integer values, please check your config")
            image_std = OmegaConf.to_container(self.config.transforms.std)
            if not isinstance(image_std, list) or any(not isinstance(x, int | float) for x in image_std):
                raise ValueError("Image std is not a list of float or integer values, please check your config")
            reports = create_mask_report(
                stage=stage,
                output=output,
                report_path="analysis_report",
                mean=image_mean,
                std=image_std,
                analysis=True,
                nb_samples=10,
                apply_sigmoid=True,
                show_orj_predictions=True,
            )
            self.metadata["report_files"].extend(reports)
            log.info("%s analysis report completed.", stage)



from __future__ import annotations

from typing import cast

import numpy as np
import pandas as pd
import torch
from numpy import ndarray
from pandas import DataFrame
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model._base import ClassifierMixin
from torch.utils.data import DataLoader

from quadra.utils import utils
from quadra.utils.classification import get_results
from quadra.utils.models import get_feature

log = utils.get_logger(__name__)


class SklearnClassificationTrainer:
    """Class to configure and run a classification using torch for feature extraction and sklearn to fit a classifier.

    Args:
        input_shape: [H, W, C]
        random_state: seed to fix randomness
        classifier: classification model
        iteration_over_training: the number of iteration over training during feature extraction
        backbone: the feature extractor
    """

    def __init__(
        self,
        input_shape: list,
        backbone: torch.nn.Module,
        random_state: int = 42,
        classifier: ClassifierMixin = LogisticRegression,
        iteration_over_training: int = 1,
    ) -> None:
        super().__init__()

        try:
            self.classifier = classifier(max_iter=1e4, random_state=random_state)
        except Exception:
            self.classifier = classifier

        self.input_shape = input_shape
        self.random_state = random_state
        self.iteration_over_training = iteration_over_training
        self.backbone = backbone

    def change_backbone(self, backbone: torch.nn.Module):
        """Update feature extractor."""
        self.backbone = backbone
        self.backbone.eval()

    def change_classifier(self, classifier: ClassifierMixin):
        """Update classifier."""
        self.classifier = classifier

    def fit(
        self,
        train_dataloader: DataLoader | None = None,
        train_features: ndarray | None = None,
        train_labels: ndarray | None = None,
    ):
        """Fit classifier on training set."""
        # Extract feature
        if self.backbone is None:
            raise AssertionError("You must set a model before running execution")

        if train_dataloader is not None:  # train_features is None or train_labels is None:
            log.info("Extracting features from training set")
            train_features, train_labels, _ = get_feature(
                feature_extractor=self.backbone,
                dl=train_dataloader,
                iteration_over_training=self.iteration_over_training,
                gradcam=False,
            )
        else:
            log.info("Using cached features for training set")
            # With the current implementation cached features are not sorted
            # Even though it doesn't seem to change anything
            if train_features is None or train_labels is None:
                raise AssertionError("Train features and labels must be provided when using cached data")
            permuted_indices = np.random.RandomState(seed=self.random_state).permutation(train_features.shape[0])
            train_features = train_features[permuted_indices]
            train_labels = train_labels[permuted_indices]

        log.info("Fitting classifier on %d features", len(train_features))  # type: ignore[arg-type]
        self.classifier.fit(train_features, train_labels)

    def test(
        self,
        test_dataloader: DataLoader,
        test_labels: ndarray | None = None,
        test_features: ndarray | None = None,
        class_to_keep: list[int] | None = None,
        idx_to_class: dict[int, str] | None = None,
        predict_proba: bool = True,
        gradcam: bool = False,
    ) -> (
        tuple[str | dict, DataFrame, float, DataFrame, np.ndarray | None]
        | tuple[None, None, None, DataFrame, np.ndarray | None]
    ):
        """Test classifier on test set.

        Args:
            test_dataloader: Test dataloader
            test_labels: test labels
            test_features: Optional test features used when cache data is available
            class_to_keep: list of class to keep
            idx_to_class: dictionary mapping class index to class name
            predict_proba: if True, predict also probability for each test image
            gradcam: Whether to compute gradcam

        Returns:
            cl_rep: Classification report
            pd_cm: Confusion matrix dataframe
            accuracy: Test accuracy
            res: Test results
            cams: Gradcams
        """
        cams = None
        # Extract feature
        if test_features is None:
            log.info("Extracting features from test set")
            test_features, final_test_labels, cams = get_feature(
                feature_extractor=self.backbone,
                dl=test_dataloader,
                gradcam=gradcam,
                classifier=self.classifier,
                input_shape=(self.input_shape[2], self.input_shape[0], self.input_shape[1]),
            )
        else:
            if test_labels is None:
                raise ValueError("Test labels must be provided when using cached data")
            log.info("Using cached features for test set")
            final_test_labels = test_labels

        # Run classifier
        log.info("Predict classifier on test set")
        test_prediction_label = self.classifier.predict(test_features)
        if predict_proba:
            test_probability = self.classifier.predict_proba(test_features)
            test_probability = test_probability.max(axis=1)

        if class_to_keep is not None:
            if idx_to_class is None:
                raise ValueError("You must provide `idx_to_class` and `test_labels` when using `class_to_keep`")
            filtered_test_labels = [int(x) if idx_to_class[x] in class_to_keep else -1 for x in final_test_labels]
        else:
            filtered_test_labels = cast(list[int], final_test_labels.tolist())

        if not hasattr(test_dataloader.dataset, "x"):
            raise ValueError("Current dataset doesn't provide an `x` attribute")

        res = pd.DataFrame(
            {
                "sample": list(test_dataloader.dataset.x),
                "real_label": final_test_labels,
                "pred_label": test_prediction_label,
            }
        )

        if not all(t == -1 for t in filtered_test_labels):
            test_real_label_cm = np.array(filtered_test_labels)
            if cams is not None:
                cams = cams[test_real_label_cm != -1]  # TODO: Is class_to_keep still used?
            pred_labels_cm = np.array(test_prediction_label)[test_real_label_cm != -1]
            test_real_label_cm = test_real_label_cm[test_real_label_cm != -1].astype(pred_labels_cm.dtype)
            cl_rep, pd_cm, accuracy = get_results(test_real_label_cm, pred_labels_cm, idx_to_class)

            if predict_proba:
                res["probability"] = test_probability

            return cl_rep, pd_cm, accuracy, res, cams

        return None, None, None, res, cams



from __future__ import annotations

import csv
import glob
import json
import os
from collections import Counter
from typing import Any, Generic, Literal, TypeVar, cast

import cv2
import hydra
import numpy as np
import torch
from anomalib.models.components.base import AnomalyModule
from anomalib.post_processing import anomaly_map_to_color_map
from anomalib.utils import plot_cumulative_histogram
from anomalib.utils.callbacks.min_max_normalization import MinMaxNormalizationCallback
from anomalib.utils.metrics.optimal_f1 import OptimalF1
from matplotlib import pyplot as plt
from omegaconf import DictConfig
from sklearn.metrics import ConfusionMatrixDisplay, f1_score
from tqdm import tqdm

from quadra.callbacks.mlflow import get_mlflow_logger
from quadra.datamodules import AnomalyDataModule
from quadra.modules.base import ModelSignatureWrapper
from quadra.tasks.base import Evaluation, LightningTask
from quadra.utils import utils
from quadra.utils.anomaly import MapOrValue, ThresholdNormalizationCallback, normalize_anomaly_score
from quadra.utils.classification import get_results
from quadra.utils.evaluation import automatic_datamodule_batch_size
from quadra.utils.export import export_model

log = utils.get_logger(__name__)

AnomalyDataModuleT = TypeVar("AnomalyDataModuleT", bound=AnomalyDataModule)


class AnomalibDetection(Generic[AnomalyDataModuleT], LightningTask[AnomalyDataModuleT]):
    """Anomaly Detection Task.

    Args:
        config: The experiment configuration
        module_function: The function that instantiates the module and model
        checkpoint_path: The path to the checkpoint to load the model from.
            Defaults to None.
        run_test: Whether to run the test after training. Defaults to False.
        report: Whether to report the results. Defaults to False.
    """

    def __init__(
        self,
        config: DictConfig,
        module_function: DictConfig,
        checkpoint_path: str | None = None,
        run_test: bool = True,
        report: bool = True,
    ):
        super().__init__(
            config=config,
            checkpoint_path=checkpoint_path,
            run_test=run_test,
            report=report,
        )
        self._module: AnomalyModule
        self.module_function = module_function
        self.export_folder = "deployment_model"
        self.report_path = ""
        self.test_results: list[dict] | None = None

    @property
    def module(self) -> AnomalyModule:
        """Get the module."""
        return self._module

    @module.setter
    def module(self, module_config):
        """Set the module."""
        if hasattr(self.config.model.model, "input_size"):
            transform_height = self.config.transforms.input_height
            transform_width = self.config.transforms.input_width
            original_model_height, original_model_width = self.config.model.model.input_size

            if transform_height != original_model_height or transform_width != original_model_width:
                log.warning(
                    "Model input size %dx%d "
                    "does not match the transform size %dx%d. "
                    "The model input size will be updated to match the transform size.",
                    original_model_height,
                    original_model_width,
                    transform_height,
                    transform_width,
                )
            self.config.model.model.input_size = [transform_height, transform_width]

        _module = cast(
            AnomalyModule,
            hydra.utils.instantiate(
                self.module_function,
                module_config,
            ),
        )

        self._module = _module

    def prepare(self) -> None:
        """Prepare the task."""
        super().prepare()
        self.module = self.config.model
        self.module.model = ModelSignatureWrapper(self.module.model)

    def export(self) -> None:
        """Export model for production."""
        if self.config.trainer.get("fast_dev_run"):
            log.warning("Skipping export since fast_dev_run is enabled")
            return

        model = self.module.model

        input_shapes = self.config.export.input_shapes

        half_precision = "16" in self.trainer.precision

        model_json, export_paths = export_model(
            config=self.config,
            model=model,
            export_folder=self.export_folder,
            half_precision=half_precision,
            input_shapes=input_shapes,
            idx_to_class={0: "good", 1: "defect"},
        )

        if len(export_paths) == 0:
            return

        model_json["image_threshold"] = np.round(self.module.image_threshold.value.item(), 3)
        model_json["pixel_threshold"] = np.round(self.module.pixel_threshold.value.item(), 3)
        model_json["anomaly_method"] = self.config.model.model.name

        with open(os.path.join(self.export_folder, "model.json"), "w") as f:
            json.dump(model_json, f, cls=utils.HydraEncoder)

    def test(self) -> Any:
        """Lightning test."""
        self.test_results = super().test()
        return self.test_results

    def _generate_report(self) -> None:
        """Generate a report for the task."""
        if len(self.report_path) > 0:
            os.makedirs(self.report_path, exist_ok=True)

        # Save json with test results
        if self.test_results is not None:
            with open(os.path.join(self.report_path, "test_results.json"), "w") as f:
                json.dump(self.test_results[0], f)

        all_output = cast(
            list[dict], self.trainer.predict(model=self.module, dataloaders=self.datamodule.test_dataloader())
        )
        all_output_flatten: dict[str, torch.Tensor | list] = {}

        for key in all_output[0]:
            if isinstance(all_output[0][key], torch.Tensor):
                tensor_gatherer = torch.cat([x[key] for x in all_output])
                all_output_flatten[key] = tensor_gatherer
            else:
                list_gatherer = []
                for x in all_output:
                    list_gatherer.extend(x[key])
                all_output_flatten[key] = list_gatherer

        image_paths = all_output_flatten["image_path"]
        named_labels = [x.split("/")[-2] for x in all_output_flatten["image_path"]]

        class_to_idx = {"good": 0}
        idx = 1
        for cls in set(named_labels):
            if cls == "good":
                continue

            class_to_idx[cls] = idx
            idx += 1

        class_to_idx["false_defect"] = idx
        idx_to_class = {v: k for k, v in class_to_idx.items()}

        gt_labels = [class_to_idx[x] for x in named_labels]
        pred_labels = []
        for i, _ in enumerate(named_labels):
            pred_label = all_output_flatten["pred_labels"][i].item()

            if pred_label == 0:
                pred_labels.append(0)
            elif pred_label == 1 and gt_labels[i] == 0:
                if idx > 2:
                    pred_labels.append(class_to_idx["false_defect"])
                else:
                    pred_labels.append(1)
            else:
                pred_labels.append(class_to_idx[named_labels[i]])

        if class_to_idx["false_defect"] not in pred_labels:
            # If there are no false defects remove the label from the confusion matrix
            class_to_idx.pop("false_defect")

        anomaly_scores = all_output_flatten["pred_scores"]

        exportable_anomaly_scores: list[Any] | np.ndarray
        if isinstance(anomaly_scores, torch.Tensor):
            exportable_anomaly_scores = anomaly_scores.cpu().numpy()
        else:
            exportable_anomaly_scores = anomaly_scores

        # Zip the lists together to create rows for the CSV file
        rows = zip(image_paths, pred_labels, gt_labels, exportable_anomaly_scores, strict=False)
        # Specify the CSV file name
        csv_file = "test_predictions.csv"
        # Write the data to the CSV file
        with open(csv_file, mode="w", newline="") as file:
            writer = csv.writer(file)
            # Write the header if needed
            writer.writerow(["image_path", "predicted_label", "ground_truth_label", "predicted_score"])
            # Write the rows
            writer.writerows(rows)

        log.info("CSV file %s has been created.", csv_file)

        if not isinstance(anomaly_scores, torch.Tensor):
            raise ValueError("Anomaly scores must be a tensor")

        good_scores = anomaly_scores[np.where(all_output_flatten["label"] == 0)]
        defect_scores = anomaly_scores[np.where(all_output_flatten["label"] == 1)]

        # Lightning has a callback attribute but is not inside the __init__ so mypy complains
        if any(
            isinstance(x, MinMaxNormalizationCallback)
            for x in self.trainer.callbacks  # type: ignore[attr-defined]
        ):
            threshold = torch.tensor(0.5)
        elif any(
            isinstance(x, ThresholdNormalizationCallback)
            for x in self.trainer.callbacks  # type: ignore[attr-defined]
        ):
            threshold = torch.tensor(100.0)
        else:
            threshold = self.module.image_metrics.F1Score.threshold

        # The output of the prediction is a normalized score so the cumulative histogram is displayed with the
        # normalized scores
        plot_cumulative_histogram(
            good_scores.cpu().numpy(), defect_scores.cpu().numpy(), threshold.item(), self.report_path
        )

        _, pd_cm, _ = get_results(np.array(gt_labels), np.array(pred_labels), idx_to_class)
        np_cm = np.array(pd_cm)
        disp = ConfusionMatrixDisplay(
            confusion_matrix=np_cm,
            display_labels=class_to_idx.keys(),
        )
        disp.plot(include_values=True, cmap=plt.cm.Greens, ax=None, colorbar=False, xticks_rotation=90)
        plt.title("Confusion Matrix")
        plt.savefig(
            os.path.join(self.report_path, "test_confusion_matrix.png"), bbox_inches="tight", pad_inches=0, dpi=300
        )
        plt.close()

        avg_score_dict = {k: 0.0 for k in set(named_labels)}

        for i, item in enumerate(named_labels):
            avg_score_dict[item] += all_output_flatten["pred_scores"][i].item()

        counter = Counter(named_labels)
        avg_score_dict = {k: v / counter[k] for k, v in avg_score_dict.items()}
        avg_score_dict = dict(sorted(avg_score_dict.items(), key=lambda q: q[1]))

        with open(os.path.join(self.report_path, "avg_score_by_label.csv"), "w") as f:
            f.write("label,avg_anomaly_score\n")
            for k, v in avg_score_dict.items():
                f.write(f"{k},{v:.3f}\n")

    def generate_report(self):
        """Generate a report for the task and try to upload artifacts."""
        self._generate_report()
        self._upload_artifacts()

    def _upload_artifacts(self):
        """If MLflow is available upload artifacts to the artifact repository."""
        mflow_logger = get_mlflow_logger(trainer=self.trainer)
        tensorboard_logger = utils.get_tensorboard_logger(trainer=self.trainer)

        if mflow_logger is not None and self.config.core.get("upload_artifacts"):
            mflow_logger.experiment.log_artifact(run_id=mflow_logger.run_id, local_path="test_confusion_matrix.png")
            mflow_logger.experiment.log_artifact(run_id=mflow_logger.run_id, local_path="avg_score_by_label.csv")

            if "visualizer" in self.config.callbacks:
                artifacts = glob.glob(os.path.join(self.config.callbacks.visualizer.output_path, "**", "*"))
                for a in artifacts:
                    mflow_logger.experiment.log_artifact(
                        run_id=mflow_logger.run_id, local_path=a, artifact_path="anomaly_output"
                    )

        if tensorboard_logger is not None and self.config.core.get("upload_artifacts"):
            artifacts = []
            artifacts.append("test_confusion_matrix.png")
            artifacts.append("avg_score_by_label.csv")

            if "visualizer" in self.config.callbacks:
                artifacts.extend(
                    glob.glob(os.path.join(self.config.callbacks.visualizer.output_path, "**/*"), recursive=True)
                )

            for a in artifacts:
                if os.path.isdir(a):
                    continue

                ext = os.path.splitext(a)[1].lower()

                if ext in [".png", ".jpg", ".jpeg", ".bmp", ".tiff", ".tif", ".gif"]:
                    try:
                        img = cv2.imread(a)
                        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                    except cv2.error:
                        log.info("Could not upload artifact image %s", a)
                        continue
                    output_path = os.path.sep.join(a.split(os.path.sep)[-2:])
                    tensorboard_logger.experiment.add_image(output_path, img, 0, dataformats="HWC")
                else:
                    utils.upload_file_tensorboard(a, tensorboard_logger)


class AnomalibEvaluation(Evaluation[AnomalyDataModule]):
    """Evaluation task for Anomalib.

    Args:
        config: Task configuration
        model_path: Path to the model folder that contains an exported model
        use_training_threshold: Whether to use the training threshold for the evaluation or use the one that
            maximizes the F1 score on the test set.
        device: Device to use for evaluation. If None, the device is automatically determined.
    """

    def __init__(
        self,
        config: DictConfig,
        model_path: str,
        use_training_threshold: bool = False,
        device: str | None = None,
        training_threshold_type: Literal["image", "pixel"] | None = None,
    ):
        super().__init__(config=config, model_path=model_path, device=device)

        self.use_training_threshold = use_training_threshold

        if training_threshold_type is not None and training_threshold_type not in ["image", "pixel"]:
            raise ValueError("Training threshold type must be either image or pixel")

        if training_threshold_type is None and use_training_threshold:
            log.warning("Using training threshold but no training threshold type is provided, defaulting to image")
            training_threshold_type = "image"

        self.training_threshold_type = training_threshold_type

    def prepare(self) -> None:
        """Prepare the evaluation."""
        super().prepare()
        self.datamodule = self.config.datamodule
        # prepare_data() must be explicitly called because there is no lightning training
        self.datamodule.prepare_data()
        self.datamodule.setup(stage="test")

    @automatic_datamodule_batch_size(batch_size_attribute_name="test_batch_size")
    def test(self) -> None:
        """Perform test."""
        log.info("Running test")
        test_dataloader = self.datamodule.test_dataloader()

        optimal_f1 = OptimalF1(num_classes=None, pos_label=1)  # type: ignore[arg-type]

        anomaly_scores = []
        anomaly_maps = []
        image_labels = []
        image_paths = []

        with torch.no_grad():
            for batch_item in tqdm(test_dataloader):
                batch_images = batch_item["image"]
                batch_labels = batch_item["label"]
                image_labels.extend(batch_labels.tolist())
                image_paths.extend(batch_item["image_path"])
                batch_images = batch_images.to(device=self.device, dtype=self.deployment_model.model_dtype)
                if self.model_data.get("anomaly_method") == "efficientad":
                    model_output = self.deployment_model(batch_images, None)
                else:
                    model_output = self.deployment_model(batch_images)
                anomaly_map, anomaly_score = model_output[0], model_output[1]
                anomaly_map = anomaly_map.cpu()
                anomaly_score = anomaly_score.cpu()
                known_labels = torch.where(batch_labels != -1)[0]
                if len(known_labels) > 0:
                    # Skip computing F1 score for images without gt
                    optimal_f1.update(anomaly_score[known_labels], batch_labels[known_labels])
                anomaly_scores.append(anomaly_score)
                anomaly_maps.append(anomaly_map)

        anomaly_scores = torch.cat(anomaly_scores)
        anomaly_maps = torch.cat(anomaly_maps)

        if any(x != -1 for x in image_labels):
            if self.use_training_threshold:
                _image_labels = torch.tensor(image_labels)
                threshold = torch.tensor(float(self.model_data[f"{self.training_threshold_type}_threshold"]))
                known_labels = torch.where(_image_labels != -1)[0]

                _image_labels = _image_labels[known_labels]
                _anomaly_scores = anomaly_scores[known_labels]

                pred_labels = (_anomaly_scores >= threshold).long()

                optimal_f1_score = torch.tensor(f1_score(_image_labels, pred_labels))
            else:
                optimal_f1_score = optimal_f1.compute()
                threshold = optimal_f1.threshold
        else:
            log.warning("No ground truth available during evaluation, use training image threshold for reporting")
            optimal_f1_score = torch.tensor(0)
            threshold = torch.tensor(float(self.model_data["image_threshold"]))

        log.info("Computed F1 score: %s", optimal_f1_score.item())
        self.metadata["anomaly_scores"] = anomaly_scores
        self.metadata["anomaly_maps"] = anomaly_maps
        self.metadata["image_labels"] = image_labels
        self.metadata["image_paths"] = image_paths
        self.metadata["threshold"] = threshold.item()
        self.metadata["optimal_f1"] = optimal_f1_score.item()

    def generate_report(self) -> None:
        """Generate report."""
        log.info("Generating report")
        if len(self.report_path) > 0:
            os.makedirs(self.report_path, exist_ok=True)

        # TODO: We currently don't use anomaly for segmentation, so the pixel threshold handling is not properly
        # implemented and we produce as output only a single threshold.
        training_threshold = self.model_data[f"{self.training_threshold_type}_threshold"]
        optimal_threshold = self.metadata["threshold"]

        normalized_optimal_threshold = cast(float, normalize_anomaly_score(optimal_threshold, training_threshold))

        os.makedirs(os.path.join(self.report_path, "predictions"), exist_ok=True)
        os.makedirs(os.path.join(self.report_path, "heatmaps"), exist_ok=True)

        anomaly_scores = self.metadata["anomaly_scores"].cpu().numpy()
        anomaly_scores = normalize_anomaly_score(anomaly_scores, training_threshold)

        if not isinstance(anomaly_scores, np.ndarray):
            raise ValueError("Anomaly scores must be a numpy array")

        good_scores = anomaly_scores[np.where(np.array(self.metadata["image_labels"]) == 0)]
        defect_scores = anomaly_scores[np.where(np.array(self.metadata["image_labels"]) == 1)]

        count_overlapping_scores = 0

        if len(good_scores) != 0 and len(defect_scores) != 0 and defect_scores.min() <= good_scores.max():
            count_overlapping_scores = len(
                np.where((anomaly_scores >= defect_scores.min()) & (anomaly_scores <= good_scores.max()))[0]
            )

        plot_cumulative_histogram(good_scores, defect_scores, normalized_optimal_threshold, self.report_path)

        json_output = {
            "observations": [],
            "threshold": np.round(normalized_optimal_threshold, 3),
            "unnormalized_threshold": np.round(optimal_threshold, 3),
            "f1_score": np.round(self.metadata["optimal_f1"], 3),
            "metrics": {
                "overlapping_scores": count_overlapping_scores,
            },
        }

        tg, fb, fg, tb = 0, 0, 0, 0

        mask_area = None
        crop_area = None

        if hasattr(self.datamodule, "valid_area_mask") and self.datamodule.valid_area_mask is not None:
            mask_area = cv2.imread(self.datamodule.valid_area_mask, 0)
            mask_area = (mask_area > 0).astype(np.uint8)

        if hasattr(self.datamodule, "crop_area") and self.datamodule.crop_area is not None:
            crop_area = self.datamodule.crop_area

        anomaly_maps = normalize_anomaly_score(self.metadata["anomaly_maps"], training_threshold)

        if not isinstance(anomaly_maps, torch.Tensor):
            raise ValueError("Anomaly maps must be a tensor")

        for img_path, gt_label, anomaly_score, anomaly_map in tqdm(
            zip(
                self.metadata["image_paths"],
                self.metadata["image_labels"],
                anomaly_scores,
                anomaly_maps,
                strict=False,
            ),
            total=len(self.metadata["image_paths"]),
        ):
            img = cv2.imread(img_path, 0)
            if mask_area is not None:
                img = img * mask_area

            if crop_area is not None:
                img = img[crop_area[1] : crop_area[3], crop_area[0] : crop_area[2]]

            output_mask = (anomaly_map >= normalized_optimal_threshold).cpu().numpy().squeeze().astype(np.uint8)
            output_mask_label = os.path.basename(os.path.dirname(img_path))
            output_mask_name = os.path.splitext(os.path.basename(img_path))[0] + ".png"
            pred_label = int(anomaly_score >= normalized_optimal_threshold)

            json_output["observations"].append(
                {
                    "image_path": os.path.dirname(img_path),
                    "file_name": os.path.basename(img_path),
                    "expectation": gt_label if gt_label != -1 else "",
                    "prediction": pred_label,
                    "prediction_mask": os.path.join("predictions", output_mask_label, output_mask_name),
                    "prediction_heatmap": os.path.join("heatmaps", output_mask_label, output_mask_name),
                    "is_correct": pred_label == gt_label if gt_label != -1 else True,
                    "anomaly_score": f"{anomaly_score.item():.3f}",
                }
            )

            if gt_label == 0 and pred_label == 0:
                tg += 1
            elif gt_label == 0 and pred_label == 1:
                fb += 1
            elif gt_label == 1 and pred_label == 0:
                fg += 1
            elif gt_label == 1 and pred_label == 1:
                tb += 1

            output_mask = output_mask * 255
            output_mask = cv2.resize(output_mask, (img.shape[1], img.shape[0]))
            output_prediction_folder = os.path.join(self.report_path, "predictions", output_mask_label)
            os.makedirs(output_prediction_folder, exist_ok=True)
            cv2.imwrite(os.path.join(output_prediction_folder, output_mask_name), output_mask)

            # Normalize the map and rescale it to 0-1 range
            # In this case we are saying that the anomaly map is in the range [normalized_th - 50, normalized_th + 50]
            # This allow to have a stronger color for the anomalies and a lighter one for really normal regions
            # It's also independent from the max or min anomaly score!
            normalized_map: MapOrValue = (anomaly_map - (normalized_optimal_threshold - 50)) / 100

            if isinstance(normalized_map, torch.Tensor):
                normalized_map = normalized_map.cpu().numpy().squeeze()

            normalized_map = np.clip(normalized_map, 0, 1)
            output_heatmap = anomaly_map_to_color_map(normalized_map, normalize=False)
            output_heatmap = cv2.resize(output_heatmap, (img.shape[1], img.shape[0]))

            output_heatmap_folder = os.path.join(self.report_path, "heatmaps", output_mask_label)
            os.makedirs(output_heatmap_folder, exist_ok=True)

            cv2.imwrite(
                os.path.join(output_heatmap_folder, output_mask_name),
                cv2.cvtColor(output_heatmap, cv2.COLOR_RGB2BGR),
            )

        json_output["metrics"]["confusion_matrix"] = {
            "class_labels": ["normal", "anomaly"],
            "matrix": [
                [tg, fb],
                [fg, tb],
            ],
        }

        with open(os.path.join(self.report_path, "anomaly_test_output.json"), "w") as f:
            json.dump(json_output, f)

    def execute(self) -> None:
        """Execute the evaluation."""
        self.prepare()
        self.test()
        self.generate_report()
        self.finalize()



import pytest

from quadra.modules.backbone import create_smp_backbone


@pytest.fixture
def smp_resnet18_unet():
    """Yield a unet with resnet18 encoder."""
    yield create_smp_backbone(
        arch="unet",
        encoder_name="resnet18",
        encoder_weights=None,
        encoder_depth=5,
        freeze_encoder=True,
        in_channels=3,
        num_classes=1,
        activation=None,
    )


@pytest.fixture
def smp_resnet18_unetplusplus():
    """Yield a unetplusplus with resnet18 encoder."""
    yield create_smp_backbone(
        arch="unetplusplus",
        encoder_name="resnet18",
        encoder_weights=None,
        encoder_depth=5,
        freeze_encoder=True,
        in_channels=3,
        num_classes=1,
        activation=None,
    )



from __future__ import annotations

import copy
import os
import random
from collections.abc import Callable, Iterable
from typing import Any

import albumentations
import matplotlib.pyplot as plt
import numpy as np
import torch
from albumentations.augmentations.transforms import Normalize
from albumentations.core.composition import TransformsSeqType
from albumentations.core.transforms_interface import NoOp
from albumentations.pytorch.transforms import ToTensorV2
from matplotlib.colors import ListedColormap
from matplotlib.lines import Line2D
from matplotlib.pyplot import get_cmap
from mpl_toolkits.axes_grid1 import ImageGrid
from omegaconf import DictConfig, ListConfig
from pytorch_grad_cam.utils.image import show_cam_on_image

from quadra.utils import utils

log = utils.get_logger(__name__)


class UnNormalize:
    """Unnormalize a tensor image with mean and standard deviation."""

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor: torch.Tensor, make_copy=True) -> torch.Tensor:
        """Call function to unnormalize a tensor image with mean and standard deviation.

        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
            make_copy (bool): whether to apply normalization to a copied tensor
        Returns:
            Tensor: Normalized image.
        """
        if make_copy:
            new_t = tensor.detach().clone()
        else:
            new_t = tensor
        for t, m, s in zip(new_t, self.mean, self.std, strict=False):
            t.mul_(s).add_(m)
            # The normalize code -> t.sub_(m).div_(s)
        return new_t


def create_grid_figure(
    images: Iterable[Iterable[np.ndarray]],
    nrows: int,
    ncols: int,
    file_path: str,
    bounds: list[tuple[float, float]],
    row_names: Iterable[str] | None = None,
    fig_size: tuple[int, int] = (12, 8),
):
    """Create a grid figure with images.

    Args:
        images: List of images to plot.
        nrows: Number of rows in the grid.
        ncols: Number of columns in the grid.
        file_path: Path to save the figure.
        row_names: Row names. Defaults to None.
        fig_size: Figure size. Defaults to (12, 8).
        bounds: Bounds for the images. Defaults to None.
    """
    default_plt_backend = plt.get_backend()
    plt.switch_backend("Agg")
    _, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=fig_size, squeeze=False)
    for i, row in enumerate(images):
        for j, image in enumerate(row):
            image_to_plot = image[0] if len(image.shape) == 3 and image.shape[0] == 1 else image
            ax[i][j].imshow(image_to_plot, vmin=bounds[i][0], vmax=bounds[i][1])
            ax[i][j].get_xaxis().set_ticks([])
            ax[i][j].get_yaxis().set_ticks([])
    if row_names is not None:
        for ax, name in zip(ax[:, 0], row_names, strict=False):  # noqa: B020
            ax.set_ylabel(name, rotation=90)

    plt.tight_layout()
    plt.savefig(file_path, bbox_inches="tight", dpi=300, facecolor="white", transparent=False)
    plt.close()
    plt.switch_backend(default_plt_backend)


def create_visualization_dataset(dataset: torch.utils.data.Dataset):
    """Create a visualization dataset by updating transforms."""

    def convert_transforms(transforms: Any):
        """Handle different types of transforms."""
        if isinstance(transforms, albumentations.BaseCompose):
            transforms.transforms = convert_transforms(transforms.transforms)
        if isinstance(transforms, list | ListConfig | TransformsSeqType):
            transforms = [convert_transforms(t) for t in transforms]
        if isinstance(transforms, dict | DictConfig):
            for tname, t in transforms.items():
                transforms[tname] = convert_transforms(t)
        if isinstance(transforms, Normalize | ToTensorV2):
            return NoOp(p=1)
        return transforms

    new_dataset = copy.deepcopy(dataset)
    # TODO: Create dataset class that has a transform attribut, we can then use isinstance
    if isinstance(dataset, torch.utils.data.Dataset):
        transform = copy.deepcopy(dataset.transform)  # type: ignore[attr-defined]
        if transform is not None:
            new_transforms = convert_transforms(transform)
            new_dataset.transform = new_transforms  # type: ignore[attr-defined]
        else:
            raise ValueError(f"The dataset transform {type(transform)} is not supported")
    else:
        raise ValueError(f"The dataset type {dataset} is not supported")
    return new_dataset


def show_mask_on_image(image: np.ndarray, mask: np.ndarray) -> np.ndarray:
    """Show a mask on an image.

    Args:
        image (np.ndarray): The image.
        mask (np.ndarray): The mask.

    Returns:
        np.ndarray: The image with the mask.
    """
    image = image.astype(np.float32) / 255
    mask = mask.astype(np.float32) / 255
    out = mask + image
    out = out / np.max(out)
    return (255 * out).astype(np.uint8)


def reconstruct_multiclass_mask(
    mask: np.ndarray,
    image_shape: tuple[int, ...],
    color_map: ListedColormap,
    ignore_class: int | None = None,
    ground_truth_mask: np.ndarray | None = None,
) -> np.ndarray:
    """Reconstruct a multiclass mask from a single channel mask.

    Args:
        mask (np.ndarray): A single channel mask.
        image_shape (Tuple[int, ...]): The shape of the image.
        color_map (ListedColormap): The color map to use.
        ignore_class (Optional[int], optional): The class to ignore. Defaults to None.
        ground_truth_mask (Optional[np.ndarray], optional): The ground truth mask. Defaults to None.

    Returns:
        mask: np.ndarray
    """
    output_mask = np.zeros(image_shape)
    for c in np.unique(mask):
        if ignore_class is not None and c == ignore_class:
            continue

        output_mask[mask == c] = color_map[str(c)]

    if ignore_class is not None and ground_truth_mask is not None:
        output_mask[ground_truth_mask == ignore_class] = [0, 0, 0]

    return output_mask


def plot_multiclass_prediction(
    image: np.ndarray,
    prediction_image: np.ndarray,
    ground_truth_image: np.ndarray,
    class_to_idx: dict[str, int],
    plot_original: bool = True,
    ignore_class: int | None = 0,
    image_height: int = 10,
    save_path: str | None = None,
    color_map: str = "tab20",
) -> None:
    """Function used to plot the image predicted.

    Args:
        image: The image to plot
        prediction_image: The prediction image
        ground_truth_image: The ground truth image
        class_to_idx: The class to idx mapping
        plot_original: Whether to plot the original image
        ignore_class: The class to ignore
        image_height: The height of the output figure
        save_path: The path to save the figure
        color_map: The color map to use. Defaults to "tab20".
    """
    image = image[0 : prediction_image.shape[0], 0 : prediction_image.shape[1], :]
    class_idxs = list(class_to_idx.values())
    cm = get_cmap(color_map)
    cmap = {str(c): tuple(int(i * 255) for i in cm(c / len(class_idxs))[:-1]) for c in class_idxs}
    output_images = []
    titles = []
    if plot_original:
        output_images.append(image)
        titles.append("Original Image")

    ground_truth_mask = reconstruct_multiclass_mask(ground_truth_image, image.shape, cmap, ignore_class=ignore_class)
    output_images.append(ground_truth_mask)
    titles.append("Ground Truth Mask")

    prediction_mask = reconstruct_multiclass_mask(
        prediction_image,
        image.shape,
        cmap,
        ignore_class=ignore_class,
    )
    output_images.append(prediction_mask)
    titles.append("Prediction Mask")
    if ignore_class is not None:
        prediction_mask = reconstruct_multiclass_mask(
            prediction_image, image.shape, cmap, ignore_class=ignore_class, ground_truth_mask=ground_truth_image
        )
        prediction_title = f"Prediction Mask \n (Ignoring Ground Truth Class: {ignore_class})"
        output_images.append(prediction_mask)
        titles.append(prediction_title)

    _, axs = plt.subplots(
        ncols=len(output_images),
        nrows=1,
        figsize=(len(output_images) * image_height, image_height),
        squeeze=False,
        facecolor="white",
    )
    for i, output_image in output_images:
        axs[0, i].imshow(show_mask_on_image(image, output_image))
        axs[0, i].set_title(titles[i])
        axs[0, i].axis("off")
    custom_lines = [Line2D([0], [0], color=tuple(i / 255.0 for i in cmap[str(c)]), lw=4) for c in class_idxs]
    custom_labels = list(class_to_idx.keys())
    axs[0, -1].legend(custom_lines, custom_labels, loc="center left", bbox_to_anchor=(1.01, 0.81), borderaxespad=0)
    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight")
        plt.close()


def plot_classification_results(
    test_dataset: torch.utils.data.Dataset,
    pred_labels: np.ndarray,
    test_labels: np.ndarray,
    class_name: str,
    original_folder: str,
    gradcam_folder: str | None = None,
    grayscale_cams: np.ndarray | None = None,
    unorm: Callable[[torch.Tensor], torch.Tensor] | None = None,
    idx_to_class: dict | None = None,
    what: str | None = None,
    real_class_to_plot: int | None = None,
    pred_class_to_plot: int | None = None,
    rows: int | None = 1,
    cols: int = 4,
    figsize: tuple[int, int] = (20, 20),
    gradcam: bool = False,
) -> None:
    """Plot and save images extracted from classification. If gradcam is True, same images
    with a gradcam heatmap (layered on original image) will also be saved.

    Args:
        test_dataset: Test dataset
        pred_labels: Predicted labels
        test_labels: Test labels
        class_name: Name of the examples' class
        original_folder: Folder where original examples will be saved
        gradcam_folder: Folder in which gradcam examples will be saved
        grayscale_cams: Grayscale gradcams (ordered as pred_labels and test_labels)
        unorm: Albumentations function to unormalize image
        idx_to_class: Dictionary of class conversion
        what: Can be "dis" or "conc", used if real_class_to_plot or pred_class_to_plot are None
        real_class_to_plot: Real class to plot.
        pred_class_to_plot: Pred class to plot.
        rows: How many rows in the plot there will be.
        cols: How many cols in the plot there will be.
        figsize: The figure size.
        gradcam: Whether to save also the gradcam version of the examples

    """
    to_plot = True
    if gradcam:
        if grayscale_cams is None:
            raise ValueError("gradcam is True but grayscale_cams is None")
        if gradcam_folder is None:
            raise ValueError("gradcam is True but gradcam_folder is None")

    if real_class_to_plot is not None:
        sample_idx = np.where(test_labels == real_class_to_plot)[0]
        if gradcam and grayscale_cams is not None:
            grayscale_cams = grayscale_cams[test_labels == real_class_to_plot]
        pred_labels = pred_labels[test_labels == real_class_to_plot]
        test_labels = test_labels[test_labels == real_class_to_plot]

    if pred_class_to_plot is not None:
        sample_idx = np.where(pred_labels == pred_class_to_plot)[0]
        if gradcam and grayscale_cams is not None:
            grayscale_cams = grayscale_cams[pred_labels == pred_class_to_plot]
        test_labels = test_labels[pred_labels == pred_class_to_plot]
        pred_labels = pred_labels[pred_labels == pred_class_to_plot]

    if pred_class_to_plot is None and real_class_to_plot is None:
        raise ValueError("'real_class_to_plot' and 'pred_class_to_plot' must not be both None")

    if what is not None:
        if what == "dis":
            cordant = pred_labels != test_labels
        elif what == "con":
            cordant = pred_labels == test_labels
        else:
            raise AssertionError(f"{what} not a valid plot type. Must be con or dis")

        sample_idx = np.array(sample_idx)[cordant]
        pred_labels = np.array(pred_labels)[cordant]
        test_labels = np.array(test_labels)[cordant]
        if gradcam:
            grayscale_cams = np.array(grayscale_cams)[cordant]

    # randomize
    idx_random = random.sample(range(len(sample_idx)), len(sample_idx))

    sample_idx = sample_idx[idx_random]
    pred_labels = pred_labels[idx_random]
    test_labels = test_labels[idx_random]
    if gradcam and grayscale_cams is not None:
        grayscale_cams = grayscale_cams[idx_random]

    cordant_chunks = list(_chunks(sample_idx, cols))

    if len(sample_idx) == 0:
        to_plot = False
        print("Nothing to plot")
    else:
        if rows is None or rows == 0:
            total_rows = len(cordant_chunks)
        else:
            total_rows = len(cordant_chunks[:rows])
        if gradcam:
            modality_list = ["original", "gradcam"]
        else:
            modality_list = ["original"]
        for modality in modality_list:
            fig = plt.figure(figsize=figsize)
            grid = ImageGrid(
                fig,
                111,  # similar to subplot(111)
                nrows_ncols=(total_rows, cols),
                axes_pad=(0.2, 0.5),
            )
            for i, ax in enumerate(grid):
                if idx_to_class is not None:
                    try:
                        pred_label = idx_to_class[pred_labels[i]]
                    except Exception:
                        pred_label = pred_labels[i]
                    try:
                        test_label = idx_to_class[test_labels[i]]
                    except Exception:
                        test_label = test_labels[i]
                else:
                    pred_label = pred_labels[i]
                    test_label = test_labels[i]

                ax.axis("off")
                ax.set_title(f"True: {str(test_label)}\nPred {str(pred_label)}")
                image, _ = test_dataset[sample_idx[i]]

                if unorm is not None:
                    image = np.array(unorm(image))
                if modality == "gradcam" and grayscale_cams is not None:
                    grayscale_cam = grayscale_cams[i]
                    rgb_cam = show_cam_on_image(
                        np.transpose(image, (1, 2, 0)), grayscale_cam, use_rgb=True, image_weight=0.7
                    )

                    ax.imshow(rgb_cam, cmap="gray")
                    if i == len(pred_labels) - 1:
                        break
                else:
                    if isinstance(image, torch.Tensor):
                        image = image.cpu().numpy()

                    if image.max() <= 1:
                        image = image * 255
                    image = image.astype(int)

                    if len(image.shape) == 3:
                        if image.shape[0] == 1:
                            image = image[0]
                        elif image.shape[0] == 3:
                            image = image.transpose((1, 2, 0))
                    ax.imshow(image, cmap="gray")
                    if i == len(pred_labels) - 1:
                        break

            for item in grid:
                item.axis("off")

            if to_plot:
                save_folder: str = ""
                if modality == "gradcam" and gradcam_folder is not None:
                    save_folder = gradcam_folder
                elif modality == "original":
                    save_folder = original_folder
                else:
                    log.warning("modality %s has no corresponding folder", modality)
                    return

                plt.savefig(
                    os.path.join(save_folder, f"{what}cordant_{class_name}_" + modality + ".png"),
                    bbox_inches="tight",
                    pad_inches=0,
                )
                plt.close()


def _chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]



from typing import Any

import segmentation_models_pytorch as smp


def create_smp_backbone(
    arch: str,
    encoder_name: str,
    freeze_encoder: bool = False,
    in_channels: int = 3,
    num_classes: int = 0,
    **kwargs: Any,
):
    """Create Segmentation.models.pytorch model backbone
    Args:
        arch: architecture name
        encoder_name: architecture name
        freeze_encoder: freeze encoder or not
        in_channels: number of input channels
        num_classes: number of classes
        **kwargs: extra arguments for model (for example classification head).
    """
    model = smp.create_model(
        arch=arch, encoder_name=encoder_name, in_channels=in_channels, classes=num_classes, **kwargs
    )
    if freeze_encoder:
        for child in model.encoder.children():
            for param in child.parameters():
                param.requires_grad = False
    return model



import os

import dotenv
from hydra.core.config_search_path import ConfigSearchPath
from hydra.plugins.search_path_plugin import SearchPathPlugin


class QuadraSearchPathPlugin(SearchPathPlugin):
    """Generic Search Path Plugin class."""

    def __init__(self):
        try:
            os.getcwd()
        except FileNotFoundError:
            # This may happen when running tests
            return

        if os.path.exists(os.path.join(os.getcwd(), ".env")):
            dotenv.load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"), override=True)

    def manipulate_search_path(self, search_path: ConfigSearchPath) -> None:
        """Plugin used to add custom config to searchpath to be discovered by quadra."""
        # This can be global or taken from the .env
        quadra_search_path = os.environ.get("QUADRA_SEARCH_PATH", None)

        # Path should be specified as a list of hydra path separated by ";"
        # E.g pkg://package1.configs;file:///path/to/configs
        if quadra_search_path is not None:
            for i, path in enumerate(quadra_search_path.split(";")):
                search_path.append(provider=f"quadra-searchpath-plugin-{i}", path=path)



from __future__ import annotations

from pathlib import Path
from typing import Any

import cv2
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
from anomalib.models.components.base import AnomalyModule
from anomalib.post_processing import (
    add_anomalous_label,
    add_normal_label,
    compute_mask,
    superimpose_anomaly_map,
)
from anomalib.pre_processing.transforms import Denormalize
from anomalib.utils.loggers import AnomalibWandbLogger
from pytorch_lightning import Callback
from pytorch_lightning.utilities.types import STEP_OUTPUT
from skimage.segmentation import mark_boundaries
from tqdm import tqdm

from quadra.utils.anomaly import MapOrValue


class Visualizer:
    """Anomaly Visualization.

    The visualizer object is responsible for collating all the images passed to it into a single image. This can then
    either be logged by accessing the `figure` attribute or can be saved directly by calling `save()` method.

    Example:
        >>> visualizer = Visualizer()
        >>> visualizer.add_image(image=image, title="Image")
        >>> visualizer.close()
    """

    def __init__(self) -> None:
        self.images: list[dict] = []

        self.figure: matplotlib.figure.Figure
        self.axis: np.ndarray

    def add_image(self, image: np.ndarray, title: str, color_map: str | None = None):
        """Add image to figure.

        Args:
          image: Image which should be added to the figure.
          title: Image title shown on the plot.
          color_map: Name of matplotlib color map used to map scalar data to colours. Defaults to None.
        """
        image_data = {"image": image, "title": title, "color_map": color_map}
        self.images.append(image_data)

    def generate(self):
        """Generate the image."""
        default_plt_backend = plt.get_backend()
        plt.switch_backend("Agg")
        num_cols = len(self.images)
        figure_size = (num_cols * 3, 3)
        self.figure, self.axis = plt.subplots(1, num_cols, figsize=figure_size)
        self.figure.subplots_adjust(right=0.9)

        axes = self.axis if len(self.images) > 1 else [self.axis]
        for axis, image_dict in zip(axes, self.images, strict=False):
            axis.axes.xaxis.set_visible(False)
            axis.axes.yaxis.set_visible(False)
            axis.imshow(image_dict["image"], image_dict["color_map"], vmin=0, vmax=255)
            axis.title.set_text(image_dict["title"])
        plt.switch_backend(default_plt_backend)

    def show(self):
        """Show image on a matplotlib figure."""
        self.figure.show()

    def save(self, filename: Path):
        """Save image.

        Args:
          filename: Filename to save image
        """
        filename.parent.mkdir(parents=True, exist_ok=True)
        self.figure.savefig(filename, dpi=100)

    def close(self):
        """Close figure."""
        plt.close(self.figure)


# TODO: This is a lot different from the 0.3.7 anomalib one
class VisualizerCallback(Callback):
    """Callback that visualizes the inference results of a model.

    The callback generates a figure showing the original image, the ground truth segmentation mask,
    the predicted error heat map, and the predicted segmentation mask.
    To save the images to the filesystem, add the 'local' keyword to the `project.log_images_to` parameter in the
    config.yaml file.

    Args:
        task: either 'segmentation' or 'classification'
        output_path: location where the images will be saved.
        inputs_are_normalized: whether the input images are normalized (like when using MinMax or Treshold callback).
        threshold_type: Either 'pixel' or 'image'. If 'pixel', the threshold is computed on the pixel-level.
        disable: whether to disable the callback.
        plot_only_wrong: whether to plot only the images that are not correctly predicted.
        plot_raw_outputs: Saves the raw images of the segmentation and heatmap output.
    """

    def __init__(
        self,
        task: str = "segmentation",
        output_path: str = "anomaly_output",
        inputs_are_normalized: bool = True,
        threshold_type: str = "pixel",
        disable: bool = False,
        plot_only_wrong: bool = False,
        plot_raw_outputs: bool = False,
    ) -> None:
        self.inputs_are_normalized = inputs_are_normalized
        self.output_path = output_path
        self.threshold_type = threshold_type
        self.disable = disable
        self.task = task
        self.plot_only_wrong = plot_only_wrong
        self.plot_raw_outputs = plot_raw_outputs

    def _add_images(self, visualizer: Visualizer, filename: Path, output_label_folder: str):
        """Save image to logger/local storage.

        Saves the image in `visualizer.figure` to the respective loggers and local storage if specified in
        `log_images_to` in `config.yaml` of the models.

        Args:
            visualizer: Visualizer object from which the `figure` is saved/logged.
            filename: Path of the input image. This name is used as name for the generated image.
            output_label_folder: ok if the image is correctly predicted or wrong if it is not
        """
        visualizer.save(
            Path(self.output_path)
            / "images"
            / output_label_folder
            / filename.parent.name
            / Path(filename.stem + ".png")
        )

    def on_test_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: AnomalyModule,
        outputs: STEP_OUTPUT | None,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Log images at the end of every batch.

        Args:
            trainer: Pytorch lightning trainer object (unused).
            pl_module: Lightning modules derived from BaseAnomalyLightning object as
                currently only they support logging images.
            outputs: Outputs of the current test step.
            batch: Input batch of the current test step (unused).
            batch_idx: Index of the current test batch (unused).
            dataloader_idx: Index of the dataloader that yielded the current batch (unused).
        """
        if self.disable:
            return

        assert outputs is not None and isinstance(outputs, dict)

        if any(x not in outputs for x in ["image_path", "image", "mask", "anomaly_maps", "label"]):
            # I'm probably in the classification scenario so I can't use the visualizer
            return

        if self.threshold_type == "pixel":
            if hasattr(pl_module.pixel_metrics.F1Score, "threshold"):
                threshold = pl_module.pixel_metrics.F1Score.threshold
            else:
                raise AttributeError("Metric has no threshold attribute")
        elif hasattr(pl_module.image_metrics.F1Score, "threshold"):
            threshold = pl_module.image_metrics.F1Score.threshold
        else:
            raise AttributeError("Metric has no threshold attribute")

        for (
            filename,
            image,
            true_mask,
            anomaly_map,
            gt_label,
            pred_label,
            anomaly_score,
        ) in tqdm(
            zip(
                outputs["image_path"],
                outputs["image"],
                outputs["mask"],
                outputs["anomaly_maps"],
                outputs["label"],
                outputs["pred_labels"],
                outputs["pred_scores"],
                strict=False,
            )
        ):
            denormalized_image = Denormalize()(image.cpu())
            current_true_mask = true_mask.cpu().numpy()
            current_anomaly_map = anomaly_map.cpu().numpy()
            # Normalize the map and rescale it to 0-1 range
            # In this case we are saying that the anomaly map is in the range [normalized_th - 50, normalized_th + 50]
            # This allow to have a stronger color for the anomalies and a lighter one for really normal regions
            # It's also independent from the max or min anomaly score!
            normalized_map: MapOrValue = (current_anomaly_map - (threshold - 50)) / 100
            normalized_map = np.clip(normalized_map, 0, 1)

            output_label_folder = "ok" if pred_label == gt_label else "wrong"

            if self.plot_only_wrong and output_label_folder == "ok":
                continue

            heatmap = superimpose_anomaly_map(
                normalized_map, denormalized_image, normalize=not self.inputs_are_normalized
            )

            if isinstance(threshold, float):
                pred_mask = compute_mask(current_anomaly_map, threshold)
            else:
                raise TypeError("Threshold should be float")
            vis_img = mark_boundaries(denormalized_image, pred_mask, color=(1, 0, 0), mode="thick")
            visualizer = Visualizer()

            if self.task == "segmentation":
                visualizer.add_image(image=denormalized_image, title="Image")
                if "mask" in outputs:
                    current_true_mask = current_true_mask * 255
                    visualizer.add_image(image=current_true_mask, color_map="gray", title="Ground Truth")
                visualizer.add_image(image=heatmap, title="Predicted Heat Map")
                visualizer.add_image(image=pred_mask, color_map="gray", title="Predicted Mask")
                visualizer.add_image(image=vis_img, title="Segmentation Result")
            elif self.task == "classification":
                gt_im = add_anomalous_label(denormalized_image) if gt_label else add_normal_label(denormalized_image)
                visualizer.add_image(gt_im, title="Image/True label")
                if anomaly_score >= threshold:
                    image_classified = add_anomalous_label(heatmap, anomaly_score)
                else:
                    image_classified = add_normal_label(heatmap, 1 - anomaly_score)
                visualizer.add_image(image=image_classified, title="Prediction")

            visualizer.generate()
            visualizer.figure.suptitle(
                f"F1 threshold: {threshold}, Mask_max: {current_anomaly_map.max():.3f}, "
                f"Anomaly_score: {anomaly_score:.3f}"
            )
            path_filename = Path(filename)
            self._add_images(visualizer, path_filename, output_label_folder)
            visualizer.close()

            if self.plot_raw_outputs:
                for raw_output, raw_name in zip([heatmap, vis_img], ["heatmap", "segmentation"], strict=False):
                    current_raw_output = raw_output
                    if raw_name == "segmentation":
                        current_raw_output = (raw_output * 255).astype(np.uint8)
                    current_raw_output = cv2.cvtColor(current_raw_output, cv2.COLOR_RGB2BGR)
                    raw_filename = (
                        Path(self.output_path)
                        / "images"
                        / output_label_folder
                        / path_filename.parent.name
                        / "raw_outputs"
                        / Path(path_filename.stem + f"_{raw_name}.png")
                    )
                    raw_filename.parent.mkdir(parents=True, exist_ok=True)
                    cv2.imwrite(str(raw_filename), current_raw_output)

    def on_test_end(self, _trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Sync logs.

        Currently only ``AnomalibWandbLogger`` is called from this method. This is because logging as a single batch
        ensures that all images appear as part of the same step.

        Args:
            _trainer: Pytorch Lightning trainer (unused)
            pl_module: Anomaly module
        """
        if self.disable:
            return

        if pl_module.logger is not None and isinstance(pl_module.logger, AnomalibWandbLogger):
            pl_module.logger.save()



from .anomaly import AnomalibDetection
from .base import Evaluation, LightningTask, PlaceholderTask, Task
from .classification import Classification, ClassificationEvaluation, SklearnClassification, SklearnTestClassification
from .patch import PatchSklearnClassification, PatchSklearnTestClassification
from .segmentation import Segmentation, SegmentationAnalysisEvaluation, SegmentationEvaluation
from .ssl import SSL

__all__ = [
    "Task",
    "LightningTask",
    "Classification",
    "ClassificationEvaluation",
    "Segmentation",
    "SegmentationEvaluation",
    "SegmentationAnalysisEvaluation",
    "SSL",
    "AnomalibDetection",
    "SklearnClassification",
    "PatchSklearnClassification",
    "PlaceholderTask",
    "Evaluation",
    "SklearnTestClassification",
    "PatchSklearnTestClassification",
]



from __future__ import annotations

import os
from ast import literal_eval
from collections.abc import Callable
from functools import wraps
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import seaborn as sns
import segmentation_models_pytorch as smp
import torch
import yaml
from segmentation_models_pytorch.losses import DiceLoss
from segmentation_models_pytorch.losses.constants import BINARY_MODE, MULTICLASS_MODE
from skimage.measure import label, regionprops  # pylint: disable=no-name-in-module

from quadra.utils.logger import get_logger
from quadra.utils.visualization import UnNormalize, create_grid_figure

try:
    from onnxruntime.capi.onnxruntime_pybind11_state import RuntimeException  # noqa

    ONNX_AVAILABLE = True
except ImportError:
    ONNX_AVAILABLE = False

log = get_logger(__name__)


def dice(
    input_tensor: torch.Tensor,
    target: torch.Tensor,
    smooth: float = 1.0,
    eps: float = 1e-8,
    reduction: str | None = "mean",
) -> torch.Tensor:
    """Dice loss computation function.

    Args:
        input_tensor:  input tensor coming from a model
        target: target tensor to compare with
        smooth: smoothing factor
        eps: epsilon to avoid zero division
        reduction: reduction method, one of "mean", "sum", "none"

    Returns:
        The computed loss
    """
    bs = input_tensor.size(0)
    iflat = input_tensor.contiguous().view(bs, -1)
    tflat = target.contiguous().view(bs, -1)
    intersection = (iflat * tflat).sum(-1)
    loss = 1 - (2.0 * intersection + smooth) / (iflat.sum(-1) + tflat.sum(-1) + smooth + eps)

    if reduction == "mean":
        loss = loss.mean()
    return loss


def score_dice(
    y_pred,
    y_true,
    reduction=None,
) -> torch.Tensor:
    """Calculate dice score."""
    return 1 - dice(y_pred, y_true, reduction=reduction)


def score_dice_smp(y_pred: torch.Tensor, y_true: torch.Tensor, mode: str = "binary") -> torch.Tensor:
    """Compute dice using smp function. Handle both binary and multiclass scenario.

    Args:
        y_pred: 1xCxHxW one channel for each class
        y_true: 1x1xHxW true mask with value in [0, ..., n_classes]
        mode: "binary" or "multiclass"

    Returns:
        dice score
    """
    if mode not in {BINARY_MODE, MULTICLASS_MODE}:
        raise ValueError(f"Mode {mode} not valid.")

    loss = DiceLoss(mode=mode, from_logits=False)

    return 1 - loss(y_pred, y_true)


def calculate_mask_based_metrics(
    images: np.ndarray,
    th_masks: torch.Tensor,
    th_preds: torch.Tensor,
    threshold: float = 0.5,
    show_orj_predictions: bool = False,
    metric: Callable = score_dice,
    multilabel: bool = False,
    n_classes: int | None = None,
) -> tuple[
    dict[str, float],
    dict[str, list[np.ndarray]],
    dict[str, list[np.ndarray]],
    dict[str, list[str | float]],
]:
    """Calculate metrics based on masks and predictions.

    Args:
        images: Images.
        th_masks: masks are tensors.
        th_preds: predictions are tensors.
        threshold: Threshold to apply. Defaults to 0.5.
        show_orj_predictions: Flag to show original predictions. Defaults to False.
        metric: Metric to use comparison. Defaults to `score_dice`.
        multilabel: True if segmentation is multiclass.
        n_classes: Number of classes. If multilabel is False, this should be None.

    Returns:
        dict: Dictionary with metrics.
    """
    masks = th_masks.cpu().numpy()
    preds = th_preds.squeeze(0).cpu().numpy()
    th_thresh_preds = (th_preds > threshold).float().cpu()
    thresh_preds = th_thresh_preds.squeeze(0).numpy()
    dice_scores = metric(th_thresh_preds, th_masks, reduction=None).numpy()
    result: dict[str, Any] = {}
    if multilabel:
        if n_classes is None:
            raise ValueError("n_classes arg shouldn't be None when multilabel is True")
        preds_multilabel = (
            torch.nn.functional.one_hot(th_preds.to(torch.int64), num_classes=n_classes).squeeze(1).permute(0, 3, 1, 2)
        )
        masks_multilabel = (
            torch.nn.functional.one_hot(th_masks.to(torch.int64), num_classes=n_classes).squeeze(1).permute(0, 3, 1, 2)
        ).to(preds_multilabel.device)
        # get_stats multiclass, not considering background channel
        tp, fp, fn, tn = smp.metrics.get_stats(
            preds_multilabel[:, 1:, :, :].long(), masks_multilabel[:, 1:, :, :].long(), mode="multilabel"
        )
    else:
        tp, fp, fn, tn = smp.metrics.get_stats(th_thresh_preds.long(), th_masks.long(), mode="binary")
    per_image_iou = smp.metrics.iou_score(tp, fp, fn, tn, reduction="micro-imagewise")
    dataset_iou = smp.metrics.iou_score(tp, fp, fn, tn, reduction="micro")
    result["F1_image"] = round(float(smp.metrics.f1_score(tp, fp, fn, tn, reduction="micro-imagewise").item()), 4)
    result["F1_pixel"] = round(float(smp.metrics.f1_score(tp, fp, fn, tn, reduction="micro").item()), 4)
    result["image_iou"] = round(float(per_image_iou.item()), 4) if not per_image_iou.isnan() else np.nan
    result["dataset_iou"] = round(float(dataset_iou.item()), 4) if not dataset_iou.isnan() else np.nan
    result["TP_pixel"] = tp.sum().item()
    result["FP_pixel"] = fp.sum().item()
    result["FN_pixel"] = fn.sum().item()
    result["TN_pixel"] = tn.sum().item()
    result["TP_image"] = 0
    result["FP_image"] = 0
    result["FN_image"] = 0
    result["TN_image"] = 0
    result["num_good_image"] = 0
    result["num_bad_image"] = 0
    bad_dice, good_dice = [], []
    fg: dict[str, list[np.ndarray]] = {"image": [], "mask": [], "thresh_pred": []}
    fb: dict[str, list[np.ndarray]] = {"image": [], "mask": [], "thresh_pred": []}
    if show_orj_predictions:
        fg["pred"] = []
        fb["pred"] = []

    area_graph: dict[str, list[str | float]] = {
        "Defect Area Percentage": [],
        "Accuracy": [],
    }
    for idx, (image, pred, mask, thresh_pred, dice_score) in enumerate(
        zip(images, preds, masks, thresh_preds, dice_scores, strict=False)
    ):
        if np.sum(mask) == 0:
            good_dice.append(dice_score)
        else:
            bad_dice.append(dice_score)
        if mask.sum() > 0:
            result["num_bad_image"] += 1
            if thresh_pred.sum() == 0:
                result["FN_image"] += 1
                fg["image"].append(image)
                fg["mask"].append(mask)
                if show_orj_predictions:
                    fg["pred"].append(pred)
                fg["thresh_pred"].append(thresh_pred)
            else:
                result["TP_image"] += 1
            rp = regionprops(label(mask[0]))
            for r in rp:
                mask_partial = th_masks[idx, :, r.bbox[0] : r.bbox[2], r.bbox[1] : r.bbox[3]]
                pred_partial = th_thresh_preds[idx, :, r.bbox[0] : r.bbox[2], r.bbox[1] : r.bbox[3]]
                tp, fp, fn, tn = smp.metrics.get_stats(pred_partial.long(), mask_partial.long(), mode="binary")
                area = tp + fn
                area_percentage = area.sum().item() * 100 / (image.shape[0] * image.shape[1])
                defect_acc = smp.metrics.accuracy(tp, fp, fn, tn, reduction="micro")
                area_graph["Accuracy"].append(defect_acc.item() * 100)
                if area_percentage <= 1:
                    area_graph["Defect Area Percentage"].append("Very Small <1%")
                elif area_percentage <= 10:
                    area_graph["Defect Area Percentage"].append("Small <10%")
                elif area_percentage <= 25:
                    area_graph["Defect Area Percentage"].append("Medium <25%")
                else:
                    area_graph["Defect Area Percentage"].append("Large >25%")

        if mask.sum() == 0:
            result["num_good_image"] += 1
            if thresh_pred.sum() > 0:
                result["FP_image"] += 1
                fb["image"].append(image)
                fb["mask"].append(mask)
                if show_orj_predictions:
                    fb["pred"].append(pred)
                fb["thresh_pred"].append(thresh_pred)
            else:
                result["TN_image"] += 1
    result["bad_dice_score_mean"] = np.mean(bad_dice) if len(bad_dice) > 0 else "null"
    result["bad_dice_score_std"] = np.std(bad_dice) if len(bad_dice) > 0 else "null"
    result["good_dice_score_mean"] = np.mean(good_dice) if len(good_dice) > 0 else "null"
    result["good_dice_score_std"] = np.std(good_dice) if len(good_dice) > 0 else "null"
    return result, fg, fb, area_graph


def create_mask_report(
    stage: str,
    output: dict[str, torch.Tensor],
    mean: npt.ArrayLike,
    std: npt.ArrayLike,
    report_path: str,
    nb_samples: int = 6,
    analysis: bool = False,
    apply_sigmoid: bool = True,
    show_all: bool = False,
    threshold: float = 0.5,
    metric: Callable = score_dice,
    show_orj_predictions: bool = False,
) -> list[str]:
    """Create report for segmentation experiment
    Args:
        stage: stage name. Train, validation or test
        output: data produced by model
        report_path: experiment path
        mean: mean values
        std: std values
        nb_samples: number of samples
        analysis: if True, analysis will be created
        apply_sigmoid: if True, sigmoid will be applied to predictions
        show_all: if True, all images will be shown
        threshold: threshold for predictions
        metric: metric function
        show_orj_predictions: if True, original predictions will be shown.

    Returns:
        list of paths to created images.
    """
    if not os.path.exists(report_path):
        os.makedirs(report_path)

    th_images = output["image"]
    th_masks = output["mask"]
    th_preds = output["mask_pred"]
    th_labels = output["label"]
    n_classes = th_preds.shape[1]
    # TODO: Apply sigmoid is a wrong name now
    # TODO: Apply sigmoid false is untested
    if apply_sigmoid:
        if n_classes == 1:
            th_preds = torch.nn.Sigmoid()(th_preds)
            th_thresh_preds = (th_preds > threshold).float()
        else:
            th_preds = torch.nn.Softmax(dim=1)(th_preds)
            th_thresh_preds = torch.argmax(th_preds, dim=1).float().unsqueeze(1)
            # Compute labels from the given masks since by default they are all 0
            th_labels = th_masks.max(dim=2)[0].max(dim=2)[0].squeeze(dim=1)
            show_orj_predictions = False
    elif n_classes == 1:
        th_thresh_preds = (th_preds > threshold).float()
    else:
        th_thresh_preds = torch.argmax(th_preds, dim=1).float().unsqueeze(1)
        # Compute labels from the given masks since by default they are all 0
        th_labels = th_masks.max(dim=2)[0].max(dim=2)[0].squeeze(dim=1)
        show_orj_predictions = False

    mean = np.asarray(mean)
    std = np.asarray(std)
    unnormalize = UnNormalize(mean, std)

    images = np.array(
        [(unnormalize(image).cpu().numpy().transpose((1, 2, 0)) * 255).astype(np.uint8) for image in th_images]
    )
    masks = th_masks.cpu().numpy()
    preds = th_preds.squeeze(0).cpu().numpy()
    thresh_preds = th_thresh_preds.squeeze(0).cpu().numpy()
    dice_scores = metric(th_thresh_preds.cpu(), th_masks.cpu(), reduction=None).numpy()

    labels = th_labels.cpu().numpy()
    binary_labels = labels == 0

    row_names = ["Input", "Mask", "Pred", f"Pred>{threshold}"]
    bounds = [(0, 255), (0.0, float(n_classes - 1)), (0.0, 1.0), (0.0, float(n_classes - 1))]
    if not show_orj_predictions:
        row_names.pop(2)
        bounds.pop(2)

    if not show_all:
        sorted_idx = np.argsort(dice_scores)
    else:
        sorted_idx = np.arange(len(dice_scores))

    binary_labels = binary_labels[sorted_idx]

    non_zero_score_idx = sorted_idx[~binary_labels]
    zero_score_idx = sorted_idx[binary_labels]
    file_paths = []
    for name, current_score_idx in zip(["good", "bad"], [zero_score_idx, non_zero_score_idx], strict=False):
        if len(current_score_idx) == 0:
            continue

        nb_total_samples = len(current_score_idx)
        nb_selected_samples = nb_total_samples if nb_samples > nb_total_samples else nb_samples
        fig_w = int(nb_selected_samples * 2)
        fig_h = int(len(row_names) * 2)
        if not show_all:
            worst_idx = current_score_idx[:nb_selected_samples].tolist()
            best_idx = current_score_idx[-nb_selected_samples:].tolist()
            random_idx = np.random.choice(current_score_idx, nb_selected_samples, replace=False).tolist()

            indexes = {"best": best_idx, "worst": worst_idx, "random": random_idx}
        else:
            indexes = {"all": current_score_idx[:nb_selected_samples].tolist()}
        for k, v in indexes.items():
            file_path = os.path.join(report_path, f"{stage}_{name}_{k}_results.png")
            images_to_show = [images[v], masks[v], preds[v], thresh_preds[v]]
            if not show_orj_predictions or n_classes > 1:
                images_to_show.pop(2)
            create_grid_figure(
                images_to_show,
                nrows=len(row_names),
                ncols=nb_selected_samples,
                row_names=row_names,
                file_path=file_path,
                fig_size=(fig_w, fig_h),
                bounds=bounds,
            )
            file_paths.append(file_path)
    if analysis:
        analysis_file_path = os.path.join(report_path, f"{stage}_analysis.yaml")
        result, fg, fb, area_graph = calculate_mask_based_metrics(
            images=images,
            th_masks=th_masks,
            th_preds=th_thresh_preds,
            threshold=threshold,
            show_orj_predictions=show_orj_predictions,
            metric=metric,
            multilabel=bool(n_classes > 1),
            n_classes=n_classes,
        )

        if len(fg["image"]) > 0:
            if len(fg["image"]) > nb_samples:
                for k, v in fg.items():
                    fg[k] = v[:nb_samples]

            fg_file_path = os.path.join(report_path, f"{stage}_fn_results.png")
            fig_w = int(len(fg["image"]) * 2)
            create_grid_figure(
                [fg for _, fg in fg.items()],
                nrows=len(row_names),
                ncols=len(fg["image"]),
                row_names=row_names,
                file_path=fg_file_path,
                fig_size=(fig_w, fig_h),
                bounds=bounds,
            )
            file_paths.append(fg_file_path)

        if len(fb["image"]) > 0:
            if len(fb["image"]) > nb_samples:
                for k, v in fb.items():
                    fb[k] = v[:nb_samples]
            fb_file_path = os.path.join(report_path, f"{stage}_fp_results.png")

            fig_w = int(len(fb["image"]) * 2)
            create_grid_figure(
                [fb for _, fb in fb.items()],
                nrows=len(row_names),
                ncols=len(fb["image"]),
                row_names=row_names,
                file_path=fb_file_path,
                fig_size=(fig_w, fig_h),
                bounds=bounds,
            )
            file_paths.append(fb_file_path)
        if len(area_graph["Defect Area Percentage"]) > 0:
            fn_area_path = os.path.join(report_path, f"{stage}_acc_area.png")
            fn_area_df = pd.DataFrame(area_graph)
            ax = sns.boxplot(
                x="Defect Area Percentage",
                y="Accuracy",
                data=fn_area_df,
                order=["Very Small <1%", "Small <10%", "Medium <25%", "Large >25%"],
            )
            ax.set_facecolor("white")
            fig = ax.get_figure()
            fig.savefig(fn_area_path)
            plt.close(fig)

            file_paths.append(fn_area_path)
        with open(analysis_file_path, "w") as file:
            yaml.dump(literal_eval(str(result)), file, default_flow_style=False)
        file_paths.append(analysis_file_path)

    return file_paths


def automatic_datamodule_batch_size(batch_size_attribute_name: str = "batch_size"):
    """Automatically scale the datamodule batch size if the given function goes out of memory.

    Args:
        batch_size_attribute_name: The name of the attribute to modify in the datamodule
    """

    def decorator(func: Callable):
        """Decorator function."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            """Wrapper function."""
            is_func_finished = False
            starting_batch_size = None
            automatic_batch_size_completed = False

            if hasattr(self, "automatic_batch_size_completed"):
                automatic_batch_size_completed = self.automatic_batch_size_completed

            if hasattr(self, "automatic_batch_size"):
                if not hasattr(self.automatic_batch_size, "disable") or not hasattr(
                    self.automatic_batch_size, "starting_batch_size"
                ):
                    raise ValueError(
                        "The automatic_batch_size attribute should have the disable and starting_batch_size attributes"
                    )
                starting_batch_size = (
                    self.automatic_batch_size.starting_batch_size if not self.automatic_batch_size.disable else None
                )

            if starting_batch_size is not None and not automatic_batch_size_completed:
                # If we already tried to reduce the batch size, we will start from the last batch size
                log.info("Performing automatic batch size scaling from %d", starting_batch_size)
                setattr(self.datamodule, batch_size_attribute_name, starting_batch_size)

            while not is_func_finished:
                valid_exceptions = (RuntimeError,)

                if ONNX_AVAILABLE:
                    valid_exceptions += (RuntimeException,)

                try:
                    func(self, *args, **kwargs)
                    is_func_finished = True
                    self.automatic_batch_size_completed = True
                    if torch.cuda.is_available():
                        torch.cuda.empty_cache()
                except valid_exceptions as e:
                    current_batch_size = getattr(self.datamodule, batch_size_attribute_name)
                    setattr(self.datamodule, batch_size_attribute_name, current_batch_size // 2)
                    log.warning(
                        "The function %s went out of memory, trying to reduce the batch size to %d",
                        func.__name__,
                        self.datamodule.batch_size,
                    )

                    if self.datamodule.batch_size == 0:
                        raise RuntimeError(
                            f"Unable to run {func.__name__} with batch size 1, the program will exit"
                        ) from e

                    if torch.cuda.is_available():
                        torch.cuda.empty_cache()

        return wrapper

    return decorator



import torch


class AsymmetricLoss(torch.nn.Module):
    """Notice - optimized version, minimizes memory allocation and gpu uploading,
    favors inplace operations.

    Args:
        gamma_neg: gamma for negative samples
        gamma_pos: gamma for positive samples
        m: bias value added to negative samples
        eps: epsilon to avoid division by zero
        disable_torch_grad_focal_loss: if True, disables torch grad for focal loss
        apply_sigmoid: if True, applies sigmoid to input before computing loss
    """

    def __init__(
        self,
        gamma_neg: float = 4,
        gamma_pos: float = 0,
        m: float = 0.05,
        eps: float = 1e-8,
        disable_torch_grad_focal_loss: bool = False,
        apply_sigmoid: bool = True,
    ):
        super().__init__()

        self.gamma_neg = gamma_neg
        self.gamma_pos = gamma_pos
        self.m = m
        self.eps = eps
        self.disable_torch_grad_focal_loss = disable_torch_grad_focal_loss
        self.apply_sigmoid = apply_sigmoid

        # prevent memory allocation and gpu uploading every iteration, and encourages inplace operations
        self.targets: torch.Tensor
        self.anti_targets: torch.Tensor
        self.xs_pos: torch.Tensor
        self.xs_neg: torch.Tensor
        self.asymmetric_w: torch.Tensor
        self.loss: torch.Tensor

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Compute the asymmetric loss.

        Args:
            x: input logits (after sigmoid)
            y: targets (multi-label binarized vector)

        Returns:
            asymettric loss
        """
        self.targets = y
        self.anti_targets = 1 - y

        # Calculating Probabilities
        self.xs_pos = x
        if self.apply_sigmoid:
            self.xs_pos = torch.sigmoid(self.xs_pos)
        self.xs_neg = 1.0 - self.xs_pos

        # Asymmetric clipping
        if self.m is not None and self.m > 0:
            self.xs_neg.add_(self.m).clamp_(max=1)

        # Basic CE calculation
        self.loss = self.targets * torch.log(self.xs_pos.clamp(min=self.eps))
        self.loss.add_(self.anti_targets * torch.log(self.xs_neg.clamp(min=self.eps)))

        # Asymmetric Focusing
        if self.gamma_neg > 0 or self.gamma_pos > 0:
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(False)
            self.xs_pos = self.xs_pos * self.targets
            self.xs_neg = self.xs_neg * self.anti_targets
            self.asymmetric_w = torch.pow(
                1 - self.xs_pos - self.xs_neg, self.gamma_pos * self.targets + self.gamma_neg * self.anti_targets
            )
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(True)
            self.loss *= self.asymmetric_w

        return -self.loss.sum()
